import logging
from pathlib import Path

from typing import Dict, List, Optional, Union

from runhouse.rns.hardware import Cluster
from runhouse.rns.packages import Package

from .env import Env


class CondaEnv(Env):
    RESOURCE_TYPE = "env"

    def __init__(
        self,
        conda_yaml: Union[str, Dict],
        name: Optional[str] = None,
        reqs: List[Union[str, Package]] = [],
        setup_cmds: List[str] = None,
        env_vars: Optional[Dict] = {},
        working_dir: Optional[Union[str, Path]] = "./",
        dryrun: bool = True,
        **kwargs,  # We have this here to ignore extra arguments when calling from_config
    ):
        """
        Runhouse CondaEnv object.

        .. note::
            To create a CondaEnv, please use the factory methods :func:`env` or :func:`conda_env`.
        """
        self.reqs = reqs
        self.conda_yaml = conda_yaml  # dict representing conda env
        super().__init__(
            name=name,
            reqs=reqs,
            setup_cmds=setup_cmds,
            env_vars=env_vars,
            working_dir=working_dir,
            dryrun=dryrun,
        )

    @staticmethod
    def from_config(config: dict, dryrun: bool = True):
        return CondaEnv(**config, dryrun=dryrun)

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        config.update({"conda_yaml": self.conda_yaml})
        return config

    @property
    def env_name(self):
        return self.conda_yaml["name"]

    def _setup_env(self, system: Cluster):
        if not ["python" in dep for dep in self.conda_yaml["dependencies"]]:
            base_python_version = system.run(["python --version"])[0][1].split()[1]
            self.conda_yaml["dependencies"].append(f"python=={base_python_version}")

        try:
            system.run(["conda --version"])
        except FileNotFoundError:
            logging.info("Conda is not installed")
            system.run(
                "wget https://repo.continuum.io/miniconda/Miniconda3-latest-Linux-x86_64.sh "
                "-O ~/miniconda.sh".split(" ")
            )
            system.run(["bash ~/miniconda.sh -b -p ~/miniconda"])
            system.run(["source $HOME/miniconda3/bin/activate"])
            status = system.run("conda --version")[0]
            if status != 0:
                raise RuntimeError("Could not install Conda.")

        path = "~/.rh/envs"
        system.run([f"mkdir -p {path}"])
        system.run_python(
            [
                "import yaml",
                "from pathlib import Path",
                f"path = Path('{path}').expanduser()",
                f"yaml.dump({self.conda_yaml}, open(path/'{self.env_name}.yml', 'w'))",
            ]
        )

        if f"\n{self.env_name} " not in system.run(["conda info --envs"])[0][1]:
            system.run([f"conda env create -f {path}/{self.env_name}.yml"])
        # TODO [CC]: throw an error if environment is not constructed correctly
        system.run(['eval "$(conda shell.bash hook)"'])
        system._sync_runhouse_to_cluster(env=self)

        if self.reqs:
            system.install_packages(self.reqs, self)

        if self.setup_cmds:
            cmd = f"{self._activate_cmd} && {self.setup_cmds.join(' && ')}"
            system.run([cmd])

    @property
    def _run_cmd(self):
        """Command prefix to run on Conda Env."""
        return f"conda run -n {self.env_name}"

    @property
    def _activate_cmd(self):
        """Command to activate Conda Env."""
        return f"conda activate {self.env_name}"



# TODO [DG] allow user to set default resource types for user, project, or org,
#  including coarse selections like 'local' or 'aws'
# TODO allow registration of user-created resource implementations



import copy
import logging
from pathlib import Path
from typing import Dict, List, Optional

import fsspec
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

import ray
import ray.data

from runhouse import Folder
from runhouse.rh_config import rns_client

from runhouse.rns.folders import folder
from runhouse.rns.resource import Resource
from runhouse.rns.utils.hardware import _current_cluster, _get_cluster_from

PREFETCH_KWARG = "prefetch_batches" if ray.__version__ >= "2.4.0" else "prefetch_blocks"

logger = logging.getLogger(__name__)


class Table(Resource):
    RESOURCE_TYPE = "table"
    DEFAULT_FOLDER_PATH = "/runhouse-table"
    DEFAULT_CACHE_FOLDER = ".cache/runhouse/tables"
    DEFAULT_STREAM_FORMAT = "pyarrow"
    DEFAULT_BATCH_SIZE = 256
    DEFAULT_PREFETCH_BATCHES = 1

    def __init__(
        self,
        path: str,
        name: Optional[str] = None,
        file_name: Optional[str] = None,
        system: Optional[str] = None,
        data_config: Optional[dict] = None,
        dryrun: bool = False,
        partition_cols: Optional[List] = None,
        stream_format: Optional[str] = None,
        metadata: Optional[Dict] = None,
        **kwargs,
    ):
        """
        The Runhouse Table object.

        .. note::
            To build a Table, please use the factory method :func:`table`.
        """
        super().__init__(name=name, dryrun=dryrun)
        self.file_name = file_name

        # Use factory method so correct subclass for system is returned
        # strip filename from path if provided
        self._folder = folder(
            path=str(Path(path).parents[0]) if Path(path).suffix else path,
            system=system,
            data_config=data_config,
            dryrun=dryrun,
        )

        self._cached_data = None
        self.partition_cols = partition_cols
        self.stream_format = stream_format or self.DEFAULT_STREAM_FORMAT
        self.metadata = metadata or {}

    @staticmethod
    def from_config(config: dict, dryrun=False):
        if isinstance(config["system"], dict):
            config["system"] = _get_cluster_from(config["system"], dryrun=dryrun)
        return _load_table_subclass(config, dryrun=dryrun)

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        if isinstance(self._folder, Resource):
            config["system"] = self._resource_string_for_subconfig(self.system)
            config["data_config"] = self._folder._data_config
        else:
            config["system"] = self.system
        self.save_attrs_to_config(
            config, ["path", "partition_cols", "metadata", "file_name"]
        )
        config.update(config)

        return config

    @property
    def data(self) -> ray.data.Dataset:
        """Get the table data. If data is not already cached, return a Ray dataset.

        With the dataset object we can stream or convert to other types, for example:

        .. code-block:: python

            data.iter_batches()
            data.to_pandas()
            data.to_dask()
        """
        if self._cached_data is not None:
            return self._cached_data
        return self._read_ray_dataset()

    @data.setter
    def data(self, new_data):
        """Update the data blob to new data"""
        self._cached_data = new_data
        # TODO should we save here?
        # self.save(overwrite=True)

    @property
    def system(self):
        return self._folder.system

    @system.setter
    def system(self, new_system):
        self._folder.system = new_system

    @property
    def path(self):
        if self.file_name:
            return f"{self._folder.path}/{self.file_name}"
        return self._folder.path

    @path.setter
    def path(self, new_path):
        self._folder.path = new_path

    def set_metadata(self, key, val):
        self.metadata[key] = val

    def get_metadata(self, key):
        return self.metadata.get(key)

    @property
    def fsspec_url(self):
        if self.file_name:
            return f"{self._folder.fsspec_url}/{self.file_name}"
        return self._folder.fsspec_url

    @property
    def data_config(self):
        return self._folder.data_config

    @data_config.setter
    def data_config(self, new_data_config):
        self._folder.data_config = new_data_config

    # @classmethod
    # def from_name(cls, name, dryrun=False):
    #     """Load existing Table via its name."""
    #     config = rns_client.load_config(name=name)
    #     if not config:
    #         raise ValueError(f"Table {name} not found.")
    #
    #     # We don't need to load the cluster dict here (if system is a cluster) because the table init
    #     # goes through the Folder factory method, which handles that.
    #
    #     # Add this table's name to the resource artifact registry if part of a run
    #     rns_client.add_upstream_resource(name)
    #
    #     # Uses the table subclass associated with the `resource_subtype`
    #     table_cls = _load_table_subclass(config=config, dryrun=dryrun)
    #     return table_cls.from_config(config=config, dryrun=dryrun)

    def to(self, system, path=None, data_config=None):
        """Copy and return the table on the given filesystem and path.

        Example:
            >>> local_table = rh.table(data, path="local/path")
            >>> s3_table = local_table.to("s3")
            >>> cluster_table = local_table.to(my_cluster)
        """
        new_table = copy.copy(self)
        new_table._folder = self._folder.to(
            system=system, path=path, data_config=data_config
        )
        return new_table

    def _save_sub_resources(self):
        if isinstance(self.system, Resource):
            self.system.save()

    def write(self):
        """Write underlying table data to fsspec URL.

        Example:
            >>> rh.table(data, path="path/to/write").write()
        """
        if self._cached_data is not None:
            data_to_write = self.data

            if isinstance(data_to_write, pd.DataFrame):
                data_to_write = self._ray_dataset_from_pandas(data_to_write)

            if isinstance(data_to_write, pa.Table):
                data_to_write = self._ray_dataset_from_arrow(data_to_write)

            if not isinstance(data_to_write, ray.data.Dataset):
                raise TypeError(f"Invalid Table format {type(data_to_write)}")

            self._write_ray_dataset(data_to_write)
            logger.info(f"Saved {str(self)} to: {self.fsspec_url}")

        return self

    def fetch(self, columns: Optional[list] = None) -> pa.Table:
        """Returns the complete table contents.

        Example:
            >>> table = rh.table(data)
            >>> fomratted_data = table.fetch()
        """
        # https://arrow.apache.org/docs/python/generated/pyarrow.parquet.read_table.html
        self._cached_data = self.read_table_from_file(columns)
        if self._cached_data is not None:
            return self._cached_data

        # When trying to read a file like object could lead to IsADirectoryError if the folder path is actually a
        # directory and the file has been automatically generated for us inside the folder
        # (ex: with pyarrow table or with partitioned data that saves multiple files within the directory)

        try:
            table_data = pq.read_table(
                self.path, columns=columns, filesystem=self._folder.fsspec_fs
            )

            if not table_data:
                raise ValueError(f"No table data found in path: {self.path}")

            self._cached_data = table_data
            return self._cached_data

        except:
            raise Exception(f"Failed to read table in path: {self.path}")

    def __getstate__(self):
        """Override the pickle method to clear _cached_data before pickling."""
        state = self.__dict__.copy()
        state["_cached_data"] = None
        return state

    def __iter__(self):
        for block in self.stream(batch_size=self.DEFAULT_BATCH_SIZE):
            for sample in block:
                yield sample

    def __len__(self):
        if isinstance(self.data, pd.DataFrame):
            len_dataset = self.data.shape[0]

        elif isinstance(self.data, ray.data.Dataset):
            len_dataset = self.data.count()

        else:
            if not hasattr(self.data, "__len__") or not self.data:
                raise RuntimeError("Cannot get len for dataset.")
            else:
                len_dataset = len(self.data)

        return len_dataset

    def __str__(self):
        return self.__class__.__name__

    def stream(
        self,
        batch_size: int,
        drop_last: bool = False,
        shuffle_seed: Optional[int] = None,
        shuffle_buffer_size: Optional[int] = None,
        prefetch_batches: Optional[int] = None,
    ):
        """Return a local batched iterator over the ray dataset.

        Example:
            >>> table = rh.table(data)
            >>> batches = table.stream(batch_size=4)
            >>> for _, batch in batches:
            >>>     print(batch)
        """
        ray_data = self.data

        if self.stream_format == "torch":
            # https://docs.ray.io/en/master/data/api/doc/ray.data.Dataset.iter_torch_batches.html#ray.data.Dataset.iter_torch_batches
            return ray_data.iter_torch_batches(
                batch_size=batch_size,
                drop_last=drop_last,
                local_shuffle_buffer_size=shuffle_buffer_size,
                local_shuffle_seed=shuffle_seed,
                # We need to do this to handle the name change of the prefetch_batches argument in ray 2.4.0
                **{PREFETCH_KWARG: prefetch_batches or self.DEFAULT_PREFETCH_BATCHES},
            )

        elif self.stream_format == "tf":
            # https://docs.ray.io/en/master/data/api/doc/ray.data.Dataset.iter_tf_batches.html
            return ray_data.iter_tf_batches(
                batch_size=batch_size,
                drop_last=drop_last,
                local_shuffle_buffer_size=shuffle_buffer_size,
                local_shuffle_seed=shuffle_seed,
                # We need to do this to handle the name change of the prefetch_batches argument in ray 2.4.0
                **{PREFETCH_KWARG: prefetch_batches or self.DEFAULT_PREFETCH_BATCHES},
            )
        else:
            # https://docs.ray.io/en/latest/data/api/dataset.html#ray.data.Dataset.iter_batches
            return ray_data.iter_batches(
                batch_size=batch_size,
                batch_format=self.stream_format,
                drop_last=drop_last,
                local_shuffle_buffer_size=shuffle_buffer_size,
                local_shuffle_seed=shuffle_seed,
                # We need to do this to handle the name change of the prefetch_batches argument in ray 2.4.0
                **{PREFETCH_KWARG: prefetch_batches or self.DEFAULT_PREFETCH_BATCHES},
            )

    def _read_ray_dataset(self, columns: Optional[List[str]] = None):
        """Read parquet data as a ray dataset object."""
        # https://docs.ray.io/en/latest/data/api/input_output.html#parquet
        dataset = ray.data.read_parquet(
            self.fsspec_url, columns=columns, filesystem=self._folder.fsspec_fs
        )
        return dataset

    def _write_ray_dataset(self, data_to_write: ray.data.Dataset):
        """Write a ray dataset to a fsspec filesystem"""
        if self.partition_cols:
            # TODO [JL]: https://arrow.apache.org/docs/python/generated/pyarrow.parquet.write_to_dataset.html
            logger.warning("Partitioning by column not currently supported.")
            pass

        # delete existing contents or they'll just be appended to
        self.rm()

        # https://docs.ray.io/en/master/data/api/doc/ray.data.Dataset.write_parquet.html
        # data_to_write.repartition(os.cpu_count() * 2).write_parquet(
        data_to_write.write_parquet(self.fsspec_url, filesystem=self._folder.fsspec_fs)

    @staticmethod
    def _ray_dataset_from_arrow(data: pa.Table):
        """Convert an Arrow Table to a Ray Dataset"""
        return ray.data.from_arrow(data)

    @staticmethod
    def _ray_dataset_from_pandas(data: pd.DataFrame):
        """Convert an Pandas DataFrame to a Ray Dataset"""
        return ray.data.from_pandas(data)

    def read_table_from_file(self, columns: Optional[list] = None):
        """Read a table from it's path.

        Example:
            >>> table = rh.table(path="path/to/table")
            >>> table_data = table.read_table_from_file()
        """
        try:
            with fsspec.open(self.fsspec_url, mode="rb", **self.data_config) as t:
                table_data = pq.read_table(t.full_name, columns=columns)
            return table_data
        except:
            return None

    def rm(self, recursive: bool = True):
        """Delete table, including its partitioned files where relevant.

        Example:
            >>> table = rh.table(path="path/to/table")
            >>> table.rm()
        """
        # https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.spec.AbstractFileSystem.rm
        self._folder.rm(recursive=recursive)

    def exists_in_system(self):
        """Whether the table exists in file system.

        Example:
            >>> table.exists_in_system()
        """
        return (
            self._folder.exists_in_system()
            and len(self._folder.ls(self.fsspec_url)) >= 1
        )


def _load_table_subclass(config: dict, dryrun: bool, data=None):
    """Load the relevant Table subclass based on the config or data type provided"""
    resource_subtype = config.get("resource_subtype", Table.__name__)

    try:
        import datasets

        if resource_subtype == "HuggingFaceTable" or isinstance(data, datasets.Dataset):
            from .huggingface_table import HuggingFaceTable

            return HuggingFaceTable(**config, dryrun=dryrun)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        if resource_subtype == "PandasTable" or isinstance(data, pd.DataFrame):
            from .pandas_table import PandasTable

            return PandasTable(**config, dryrun=dryrun)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        import dask.dataframe as dd

        if resource_subtype == "DaskTable" or isinstance(data, dd.DataFrame):
            from .dask_table import DaskTable

            return DaskTable(**config, dryrun=dryrun)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        import ray

        if resource_subtype == "RayTable" or isinstance(data, ray.data.Dataset):
            from .ray_table import RayTable

            return RayTable(**config, dryrun=dryrun)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        import cudf

        if resource_subtype == "CudfTable" or isinstance(data, cudf.DataFrame):
            raise NotImplementedError("Cudf not currently supported")
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    if resource_subtype == "Table" or isinstance(data, pa.Table):
        new_table = Table(**config, dryrun=dryrun)
        return new_table
    else:
        raise TypeError(
            f"Unsupported data type {type(data)} for Table construction. "
            f"For converting data to pyarrow see: "
            f"https://arrow.apache.org/docs/7.0/python/generated/pyarrow.Table.html"
        )


def table(
    data=None,
    name: Optional[str] = None,
    path: Optional[str] = None,
    system: Optional[str] = None,
    data_config: Optional[dict] = None,
    partition_cols: Optional[list] = None,
    mkdir: bool = False,
    dryrun: bool = False,
    stream_format: Optional[str] = None,
    metadata: Optional[dict] = None,
) -> Table:
    """Constructs a Table object, which can be used to interact with the table at the given path.

    Args:
        data: Data to be stored in the table.
        name (Optional[str]): Name for the table, to reuse it later on.
        path (Optional[str]): Full path to the data file.
        system (Optional[str]): File system. Currently this must be one of:
            [``file``, ``github``, ``sftp``, ``ssh``, ``s3``, ``gs``, ``azure``].
        data_config (Optional[dict]): The data config to pass to the underlying fsspec handler.
        partition_cols (Optional[list]): List of columns to partition the table by.
        mkdir (bool): Whether to create a remote folder for the table. (Default: ``False``)
        dryrun (bool): Whether to create the Table if it doesn't exist, or load a Table object as a dryrun.
            (Default: ``False``)
        stream_format (Optional[str]): Format to stream the Table as.
            Currently this must be one of: [``pyarrow``, ``torch``, ``tf``, ``pandas``]
        metadata (Optional[dict]): Metadata to store for the table.

    Returns:
        Table: The resulting Table object.

    Example:
        >>> import runhouse as rh
        >>> # Create and save (pandas) table
        >>> rh.table(
        >>>    data=data,
        >>>    name="~/my_test_pandas_table",
        >>>    path="table_tests/test_pandas_table.parquet",
        >>>    system="file",
        >>>    mkdir=True,
        >>> ).save()
        >>>
        >>> # Load table from above
        >>> reloaded_table = rh.table(name="~/my_test_pandas_table")
    """
    if name and not any(
        [
            data is not None,
            path,
            system,
            data_config,
            partition_cols,
            stream_format,
            metadata,
        ]
    ):
        # Try reloading existing table
        return Table.from_name(name, dryrun)

    system = _get_cluster_from(
        system or _current_cluster(key="config") or Folder.DEFAULT_FS, dryrun=dryrun
    )

    file_name = None
    if path:
        # Extract the file name from the path if provided
        full_path = Path(path)
        file_suffix = full_path.suffix
        if file_suffix:
            path = str(full_path.parent)
            file_name = full_path.name

    if path is None:
        # If no path is provided we need to create one based on the name of the table
        table_name_in_path = rns_client.resolve_rns_data_resource_name(name)
        if system == rns_client.DEFAULT_FS or (
            isinstance(system, Resource) and system.on_this_cluster()
        ):
            # create random path to store in .cache folder of local filesystem
            path = str(
                Path(
                    f"~/{Table.DEFAULT_CACHE_FOLDER}/{table_name_in_path}"
                ).expanduser()
            )
        else:
            # save to the default bucket
            path = f"{Table.DEFAULT_FOLDER_PATH}/{table_name_in_path}"

    config = {
        "system": system,
        "name": name,
        "path": path,
        "file_name": file_name,
        "data_config": data_config,
        "partition_cols": partition_cols,
        "stream_format": stream_format,
        "metadata": metadata,
    }

    new_table = _load_table_subclass(config=config, dryrun=dryrun, data=data)
    if data is not None:
        new_table.data = data

    return new_table



from runhouse.rns.top_level_rns_fns import (
    current_folder,
    exists,
    get_pinned_object,
    ipython,
    load,
    locate,
    pin_to_memory,
    resources,
    set_folder,
    unset_folder,
)

# Note these are global variables that are instantiated within rh_config.py:
from .rh_config import configs, obj_store, rns_client
from .rns.blob import blob, Blob
from .rns.defaults import Defaults
from .rns.envs import conda_env, CondaEnv, env, Env
from .rns.folders import Folder, folder, GCSFolder, S3Folder
from .rns.function import function, Function
from .rns.hardware import cluster, Cluster, ondemand_cluster, OnDemandCluster
from .rns.kvstores.kvstore import KVStore
from .rns.login import login, logout
from .rns.packages import git_package, GitPackage, package, Package
from .rns.run import Run, run, RunStatus, RunType
from .rns.secrets.secrets import Secrets
from .rns.tables.table import table, Table

# Briefly keep for BC.
send = function

# Syntactic sugar
fn = function

__version__ = "0.0.8"



import copy
import unittest

import pytest

import runhouse as rh

from runhouse.rns.hardware import OnDemandCluster


def is_on_cluster(cluster):
    return cluster.on_this_cluster()


@pytest.mark.clustertest
def test_cluster_config(cpu_cluster):
    config = cpu_cluster.config_for_rns
    cluster2 = OnDemandCluster.from_config(config)
    assert cluster2.address == cpu_cluster.address


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_cluster_sharing(cpu_cluster):
    cpu_cluster.share(
        users=["donny@run.house", "josh@run.house"],
        access_type="write",
        notify_users=False,
    )
    assert True


@pytest.mark.clustertest
def test_read_shared_cluster(cpu_cluster):
    res = cpu_cluster.run_python(["import numpy", "print(numpy.__version__)"])
    assert res[0][1]


@pytest.mark.clustertest
def test_install(cpu_cluster):
    cpu_cluster.install_packages(
        [
            "./",
            "torch==1.12.1",
            # 'conda:jupyterlab',
            # 'gh:pytorch/vision'
        ]
    )


@pytest.mark.clustertest
def test_basic_run(cpu_cluster):
    # Create temp file where fn's will be stored
    test_cmd = "echo hi"
    cpu_cluster.up_if_not()
    res = cpu_cluster.run(commands=[test_cmd])
    assert "hi" in res[0][1]


@pytest.mark.clustertest
def test_restart_server(cpu_cluster):
    cpu_cluster.up_if_not()
    codes = cpu_cluster.restart_server(resync_rh=False)
    assert codes


@pytest.mark.clustertest
def test_on_same_cluster(cpu_cluster):
    hw_copy = copy.copy(cpu_cluster)
    cpu_cluster.up_if_not()

    func_hw = rh.function(is_on_cluster).to(cpu_cluster)
    assert func_hw(cpu_cluster)
    assert func_hw(hw_copy)


@pytest.mark.clustertest
def test_on_diff_cluster(cpu_cluster, byo_cpu):
    func_hw = rh.function(is_on_cluster).to(cpu_cluster)
    assert not func_hw(byo_cpu)


if __name__ == "__main__":
    unittest.main()



import re
import subprocess

from pathlib import Path
from typing import Dict, List

import yaml

from runhouse.rh_config import rns_client
from runhouse.rns.packages import Package
from runhouse.rns.resource import Resource


def _process_reqs(reqs):
    preprocessed_reqs = []
    for package in reqs:
        # TODO [DG] the following is wrong. RNS address doesn't have to start with '/'. However if we check if each
        #  string exists in RNS this will be incredibly slow, so leave it for now.
        if (
            isinstance(package, str)
            and package[0] == "/"
            and rns_client.exists(package)
        ):
            # If package is an rns address
            package = rns_client.load_config(package)
        elif (
            isinstance(package, str)
            and Path(package.split(":")[-1]).expanduser().exists()
        ):
            # if package refers to a local path package
            package = Package.from_string(package)
        elif isinstance(package, dict):
            package = Package.from_config(package)
        preprocessed_reqs.append(package)
    return preprocessed_reqs


def _get_env_from(env):
    if isinstance(env, Resource):
        return env

    from runhouse.rns.envs import Env

    if isinstance(env, List):
        return Env(reqs=env, working_dir="./")
    elif isinstance(env, Dict):
        return Env.from_config(env)
    elif isinstance(env, str) and rns_client.exists(env, resource_type="env"):
        return Env.from_name(env)
    return env


def _get_conda_yaml(conda_env=None):
    if not conda_env:
        return None
    if isinstance(conda_env, str):
        if Path(conda_env).expanduser().exists():  # local yaml path
            conda_yaml = yaml.safe_load(open(conda_env))
        elif f"\n{conda_env} " in subprocess.check_output(
            "conda info --envs".split(" ")
        ).decode("utf-8"):
            res = subprocess.check_output(
                f"conda env export -n {conda_env} --no-build".split(" ")
            ).decode("utf-8")
            conda_yaml = yaml.load(res)
        else:
            raise Exception(
                f"{conda_env} must be a Dict or point to an existing path or conda environment."
            )
    else:
        conda_yaml = conda_env

    # ensure correct version to Ray -- this is subject to change if SkyPilot adds additional ray version support
    conda_yaml["dependencies"] = (
        conda_yaml["dependencies"] if "dependencies" in conda_yaml else []
    )
    if not [dep for dep in conda_yaml["dependencies"] if "pip" in dep]:
        conda_yaml["dependencies"].append("pip")
    if not [
        dep
        for dep in conda_yaml["dependencies"]
        if isinstance(dep, Dict) and "pip" in dep
    ]:
        conda_yaml["dependencies"].append({"pip": ["ray<=2.4.0,>=2.2.0"]})
    else:
        for dep in conda_yaml["dependencies"]:
            if (
                isinstance(dep, Dict)
                and "pip" in dep
                and not [pip for pip in dep["pip"] if "ray" in pip]
            ):
                dep["pip"].append("ray<=2.4.0,>=2.2.0")
                continue
    return conda_yaml


def _env_vars_from_file(env_file):
    env_file = Path(env_file) if isinstance(env_file, str) else env_file
    if not env_file.exists():
        raise FileNotFoundError(f"Can not find provided env file: {env_file}")

    env_vars_dict = {}
    with open(env_file) as f:
        for line in f:
            line = line.strip()
            if line and not line.startswith("#") and "=" in line:
                key, value = re.split("\s*=\s*", line, maxsplit=1)
                env_vars_dict[key] = value

    return env_vars_dict



import subprocess
import unittest
from pathlib import Path

import pytest

import runhouse as rh

extra_index_url = "--extra-index-url https://pypi.python.org/simple/"
cuda_116_url = "--index-url https://download.pytorch.org/whl/cu116"


def summer(a, b):
    return a + b


def send_tensor_to_cuda():
    try:
        import torch

        a = torch.LongTensor(1).random_(0, 10)
        a = a.to(device="cuda")
        return isinstance(a, torch.Tensor)

    except Exception as e:
        return e


@pytest.mark.localtest
def test_from_string():
    p = rh.Package.from_string("reqs:~/runhouse")
    if (Path.home() / "runhouse").exists():
        assert isinstance(p.install_target, rh.Folder)
        assert p.install_target.path == str(Path.home() / "runhouse")
    else:
        assert p.install_target == "~/runhouse"


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_share_package(cpu_cluster, local_package):
    local_package.to(system=cpu_cluster)
    local_package.save("package_to_share")  # shareable resource requires a name

    local_package.share(
        users=["josh@run.house", "donny@run.house"],
        access_type="write",
        notify_users=False,
    )

    # TODO test loading from a different account for real
    # Confirm the package's folder is now on the cluster
    status_codes = cpu_cluster.run(commands=["ls tmp_package"])
    assert "sample_file_0.txt" in status_codes[0][1]


@pytest.mark.rnstest
def test_share_git_package():
    git_package = rh.GitPackage(
        name="shared_git_package",
        git_url="https://github.com/runhouse/runhouse.git",
        install_method="pip",
        revision="v0.0.1",
    ).save()

    git_package.share(
        users=["donny@run.house", "josh@run.house"],
        access_type="read",
        notify_users=False,
    )
    assert rh.rns_client.exists(name_or_path="shared_git_package")


@pytest.mark.rnstest
def test_load_shared_git_package():
    git_package = rh.package(name="@/shared_git_package")
    assert git_package.config_for_rns


@pytest.mark.clustertest
def test_local_package_function(cpu_cluster):
    function = rh.function(fn=summer).to(cpu_cluster, env=["./"])

    req = function.env.reqs[0]
    assert isinstance(req, rh.Package)
    assert isinstance(req.install_target, rh.Folder)
    assert req.install_target.system == cpu_cluster


@pytest.mark.clustertest
def test_local_package_to_cluster(cpu_cluster):
    package = rh.Package.from_string("./").to(cpu_cluster)

    assert isinstance(package.install_target, rh.Folder)
    assert package.install_target.system == cpu_cluster


@pytest.mark.clustertest
def test_mount_local_package_to_cluster(cpu_cluster):
    mount_path = "package_mount"
    package = rh.Package.from_string("./").to(cpu_cluster, path=mount_path, mount=True)

    assert isinstance(package.install_target, rh.Folder)
    assert package.install_target.system == cpu_cluster
    assert mount_path in cpu_cluster.run(["ls"])[0][1]


@pytest.mark.clustertest
@pytest.mark.awstest
def test_package_file_system_to_cluster(cpu_cluster, s3_package):
    assert s3_package.install_target.system == "s3"
    assert s3_package.install_target.exists_in_system()

    folder_name = Path(s3_package.install_target.path).stem
    s3_package.to(system=cpu_cluster, mount=True, path=folder_name)

    # Confirm the package's folder is now on the cluster
    assert "sample_file_0.txt" in cpu_cluster.run([f"ls {folder_name}"])[0][1]


@pytest.mark.localtest
@pytest.mark.parametrize(
    "reqs_lines",
    [
        [cuda_116_url, "", "torch"],
        ["diffusers", "accelerate"],
        [f"torch {cuda_116_url}"],
    ],
)
def test_basic_command_generator_from_reqs(reqs_lines):
    test_reqs_file = Path(__file__).parent / "requirements.txt"
    with open(test_reqs_file, "w") as f:
        f.writelines([line + "\n" for line in reqs_lines])

    dummy_pkg = rh.Package.from_string(specifier="pip:dummy_package")
    install_cmd = dummy_pkg._requirements_txt_install_cmd(test_reqs_file)

    assert install_cmd == f"-r {test_reqs_file}"

    test_reqs_file.unlink()
    assert True


@pytest.mark.localtest
def test_command_generator_from_reqs():
    reqs_lines = ["torch", "accelerate"]
    test_reqs_file = Path(__file__).parent / "requirements.txt"
    with open(test_reqs_file, "w") as f:
        f.writelines([line + "\n" for line in reqs_lines])

    dummy_pkg = rh.Package.from_string(specifier="pip:dummy_package")
    install_cmd = dummy_pkg._requirements_txt_install_cmd(
        test_reqs_file, cuda_version_or_cpu="11.6"
    )

    assert (
        install_cmd
        == f"-r {test_reqs_file} --extra-index-url https://download.pytorch.org/whl/cu116"
    )

    test_reqs_file.unlink()
    assert True


@pytest.mark.localtest
def test_torch_install_command_generator_from_reqs():
    """Test correctly generating full install commands for torch-related packages."""
    test_cuda_version = "11.6"

    # [Required as listed in reqs.txt, expected formatted install cmd]
    packages_to_install = [
        [f"torch {extra_index_url}", f"torch {cuda_116_url} {extra_index_url}"],
        ["torchaudio", f"torchaudio {cuda_116_url} {extra_index_url}"],
        ["matchatorch", f"matchatorch {cuda_116_url} {extra_index_url}"],
        ["torch==1.13.0", f"torch==1.13.0 {cuda_116_url} {extra_index_url}"],
        [
            f"torch {cuda_116_url} {extra_index_url}",
            f"torch {cuda_116_url} {extra_index_url}",
        ],
        [
            f"torch>=1.13.0 {cuda_116_url}",
            f"torch>=1.13.0 {cuda_116_url} {extra_index_url}",
        ],
    ]

    dummy_pkg = rh.Package.from_string(specifier="pip:dummy_package")
    reformatted_packaged_to_install = [
        dummy_pkg._install_cmd_for_torch(p[0], test_cuda_version)
        for p in packages_to_install
    ]

    for idx, install_cmd in enumerate(reformatted_packaged_to_install):
        assert install_cmd == packages_to_install[idx][1]


@pytest.mark.localtest
def test_torch_install_command_generator():
    """Checks that the command itself is correct (without actually running it on the cluster)"""
    cuda_version = "11.6"
    # [Mock command received by user, cuda version, expected formatted command]
    packages_to_install = [
        [
            f"torch==1.13.1 {extra_index_url}",
            cuda_version,
            f"torch==1.13.1 {cuda_116_url} {extra_index_url}",
        ],
        [
            f"torch {cuda_116_url} {extra_index_url}",
            cuda_version,
            f"torch {cuda_116_url} {extra_index_url}",
        ],
        ["matchatorch", cuda_version, f"matchatorch {cuda_116_url} {extra_index_url}"],
        [
            "torch>=1.13.0, <2.0.0",
            cuda_version,
            f"torch>=1.13.0,<2.0.0 {cuda_116_url} {extra_index_url}",
        ],
        [
            f"torch {cuda_116_url} torchaudio {extra_index_url}",
            cuda_version,
            f"torch {cuda_116_url} {extra_index_url} "
            f"torchaudio {cuda_116_url} "
            f"{extra_index_url}",
        ],
        [
            "torch torchpudding",
            cuda_version,
            f"torch {cuda_116_url} {extra_index_url} torchpudding {cuda_116_url} {extra_index_url}",
        ],
        [
            "torch>=1.13.0",
            cuda_version,
            f"torch>=1.13.0 {cuda_116_url} {extra_index_url}",
        ],
        [
            f"torch>1.13.0 {extra_index_url}",
            cuda_version,
            f"torch>1.13.0 {cuda_116_url} {extra_index_url}",
        ],
        [
            f"torch==1.13.0 {cuda_116_url}",
            cuda_version,
            f"torch==1.13.0 {cuda_116_url} {extra_index_url}",
        ],
        [
            f"torch==1.13.0 {cuda_116_url}",
            cuda_version,
            f"torch==1.13.0 {cuda_116_url} {extra_index_url}",
        ],
        [
            "torch~=1.13.0",
            cuda_version,
            f"torch~=1.13.0 {cuda_116_url} {extra_index_url}",
        ],
        [
            f"torch==99.99.999 {cuda_116_url}",
            cuda_version,
            f"torch==99.99.999 {cuda_116_url} {extra_index_url}",
        ],
    ]
    for cmds in packages_to_install:
        torch_version, cuda_version, expected_install_cmd = cmds
        dummy_pkg = rh.Package.from_string(specifier=f"pip:{torch_version}")
        formatted_install_cmd = dummy_pkg._install_cmd_for_torch(
            torch_version, cuda_version
        )

        assert (
            formatted_install_cmd == expected_install_cmd
        ), f"Unexpected result for {(torch_version, cuda_version)} "


@pytest.mark.gputest
@pytest.mark.parametrize(
    "cluster",
    ["cpu_cluster", "v100_gpu_cluster", "k80_gpu_cluster", "a10g_gpu_cluster"],
    indirect=True,
)
def test_getting_cuda_version_on_clusters(request, cluster):
    """Gets the cuda version on the cluster and asserts it is the expected version"""
    return_codes: list = cluster.run_python(
        ["import runhouse as rh", "print(rh.Package._detect_cuda_version_or_cpu())"]
    )
    cuda_version_or_cpu = return_codes[0][1].strip().split("\n")[-1]
    print(f"{cluster.name}: {cuda_version_or_cpu}")

    instance_type = cluster.instance_type.lower()

    # Save the cuda version (or indicate cpu) for each cluster
    request.config.cache.set(instance_type, cuda_version_or_cpu)

    if instance_type.startswith("k80"):
        assert cuda_version_or_cpu == "11.3"
    elif instance_type.startswith("cpu"):
        assert cuda_version_or_cpu == "cpu"
    else:
        assert cuda_version_or_cpu == "11.7"


@pytest.mark.gputest
@pytest.mark.parametrize(
    "cluster",
    ["cpu_cluster", "v100_gpu_cluster", "k80_gpu_cluster", "a10g_gpu_cluster"],
    indirect=True,
)
def test_install_cmd_for_torch_on_cluster(request, cluster):
    """Checks that the install command for torch runs properly on the cluster.
    Confirms that we can properly install the package (and send a torch tensor to cuda to validate it"""
    # Use the cuda version on the cluster that we got from the previous test
    cuda_version_or_cpu = request.config.cache.get(cluster.instance_type.lower(), None)

    assert "AttributeError" not in cuda_version_or_cpu, (
        f"No cuda version saved for {cluster.name} - run `test_getting_cuda_version_on_clusters`"
        "to save the cuda version"
    )

    cuda_url = rh.Package.TORCH_INDEX_URLS.get(cuda_version_or_cpu)
    cuda_index_url = f"--index-url {cuda_url}"

    install_commands_for_cluster = [
        f"torch==1.13.1 {cuda_index_url} {extra_index_url}",
        f"torch>=1.13.0 {extra_index_url}",
        f"torch>1.13.0 {cuda_index_url}",
        f"torch {cuda_index_url}",
        f"torch>=1.13.0,<=2.0.0 {cuda_index_url}",
        f"torch==2.0.0 {cuda_index_url} torchaudio {cuda_index_url}",
        f"torch~=2.0.0 {cuda_index_url}",
    ]

    for install_cmd in install_commands_for_cluster:
        # Run the complete install command on the cluster
        try:
            cluster.install_packages([install_cmd])
        except subprocess.CalledProcessError:
            assert False, f"Failed to install {install_cmd}"

    if cuda_version_or_cpu != "cpu":
        # Send a tensor to CUDA using the installed torch package (ignore if we are on CPU)
        tensor_to_cuda = rh.function(fn=send_tensor_to_cuda).to(
            cluster, reqs=["pytest"]
        )
        sent_to_cuda = tensor_to_cuda()
        assert sent_to_cuda, f"Failed to send torch tensor to CUDA on {cluster.name}"


if __name__ == "__main__":
    unittest.main()



import os
import time
import unittest
from multiprocessing import Pool

import pytest
import ray.exceptions
import requests
import runhouse as rh
from runhouse.rns.api_utils.resource_access import ResourceAccess
from runhouse.rns.api_utils.utils import load_resp_content

REMOTE_FUNC_NAME = "@/remote_function"


def call_function(fn, **kwargs):
    return fn(**kwargs)


def torch_summer(a, b):
    # import inside so tests that don't use torch don't fail because torch isn't in their reqs
    import torch

    return int(torch.Tensor([a, b]).sum())


def summer(a, b):
    return a + b


def np_array(list):
    import numpy as np

    return np.array(list)


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_create_function_from_name_local(cpu_cluster):
    local_name = "~/local_function"
    local_sum = rh.function(summer).to(cpu_cluster).save(local_name)
    del local_sum

    remote_sum = rh.function(name=local_name)
    res = remote_sum(1, 5)
    assert res == 6

    remote_sum.delete_configs()
    assert rh.exists(local_name) is False


@unittest.skip("Not yet implemented.")
@pytest.mark.rnstest
@pytest.mark.clustertest
def test_running_function_as_proxy(cpu_cluster):
    # reload the function from RNS
    remote_sum = rh.function(name=REMOTE_FUNC_NAME)
    remote_sum.access = ResourceAccess.PROXY
    res = remote_sum(1, 5)
    assert res == 6

    remote_sum.delete_configs()
    assert not rh.exists(REMOTE_FUNC_NAME)


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_create_function_from_rns(cpu_cluster):
    remote_sum = rh.function(summer).to(cpu_cluster).save(REMOTE_FUNC_NAME)
    del remote_sum

    # reload the function
    remote_sum = rh.function(name=REMOTE_FUNC_NAME)
    res = remote_sum(1, 5)
    assert res == 6

    remote_sum.delete_configs()
    assert not rh.exists(REMOTE_FUNC_NAME)


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_get_function_history(cpu_cluster):
    # reload the function from RNS
    remote_sum = rh.function(summer).to(cpu_cluster).save(REMOTE_FUNC_NAME)

    history = remote_sum.history()
    assert history


def multiproc_torch_sum(inputs):
    print(f"CPUs: {os.cpu_count()}")
    with Pool() as P:
        return P.starmap(torch_summer, inputs)


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_remote_function_with_multiprocessing(cpu_cluster):
    re_fn = rh.function(multiproc_torch_sum, name="test_function").to(
        cpu_cluster, env=["torch==1.12.1"]
    )

    summands = list(zip(range(5), range(4, 9)))
    res = re_fn(summands)

    assert res == [4, 6, 8, 10, 12]


def getpid(a=0):
    return os.getpid() + a


@pytest.mark.clustertest
def test_maps(cpu_cluster):
    pid_fn = rh.function(getpid, system=cpu_cluster)
    num_pids = [1] * 10
    pids = pid_fn.map(num_pids)
    assert len(set(pids)) > 1
    assert all(pid > 0 for pid in pids)

    pids = pid_fn.repeat(num_repeats=10)
    assert len(set(pids)) > 1
    assert all(pid > 0 for pid in pids)

    pids = [pid_fn.enqueue() for _ in range(10)]
    assert len(pids) == 10
    assert all(pid > 0 for pid in pids)

    re_fn = rh.function(summer, system=cpu_cluster)
    summands = list(zip(range(5), range(4, 9)))
    res = re_fn.starmap(summands)
    assert res == [4, 6, 8, 10, 12]

    alist, blist = range(5), range(4, 9)
    res = re_fn.map(alist, blist)
    assert res == [4, 6, 8, 10, 12]


@pytest.mark.clustertest
def test_remotes(cpu_cluster):
    pid_fn = rh.function(getpid, system=cpu_cluster)

    pid_run = pid_fn.run()
    pid_res = pid_fn.get(pid_run.name)
    assert pid_res > 0

    # Test passing an objectref into a normal call
    pid_res_from_ref = pid_fn(pid_run.name)
    assert pid_res_from_ref > pid_res


@pytest.mark.clustertest
def test_function_git_fn(cpu_cluster):
    remote_parse = rh.function(
        fn="https://github.com/huggingface/diffusers/blob/"
        "main/examples/dreambooth/train_dreambooth.py:parse_args",
        system=cpu_cluster,
        env=[
            "torch==1.12.1 --verbose",
            "torchvision==0.13.1",
            "transformers",
            "datasets",
            "evaluate",
            "accelerate",
            "pip:./diffusers --verbose",
        ],
    )
    args = remote_parse(
        input_args=[
            "--pretrained_model_name_or_path",
            "stabilityai/stable-diffusion-2-base",
            "--instance_data_dir",
            "remote_image_dir",
            "--instance_prompt",
            "a photo of sks person",
        ]
    )
    assert args.pretrained_model_name_or_path == "stabilityai/stable-diffusion-2-base"


@pytest.mark.clustertest
def test_list_keys(cpu_cluster):
    pid_fn = rh.function(getpid).to(system=cpu_cluster)

    pid_obj1 = pid_fn.run()
    pid_obj2 = pid_fn.run()

    current_jobs = cpu_cluster.list_keys()
    assert set([pid_obj1.name, pid_obj2.name]).issubset(current_jobs)


def slow_getpid(a=0):
    time.sleep(10)
    return os.getpid() + a


@pytest.mark.clustertest
def test_cancel_jobs(cpu_cluster):
    pid_fn = rh.function(slow_getpid).to(cpu_cluster)

    pid_run1 = pid_fn.run(2)

    time.sleep(1)  # So the runkeys are more than 1 second apart
    pid_ref2 = pid_fn.run(5)

    print("Cancelling jobs")
    cpu_cluster.cancel_all()

    with pytest.raises(Exception) as e:
        print(pid_fn.get(pid_run1.name))
        assert isinstance(
            e, (ray.exceptions.TaskCancelledError, ray.exceptions.RayTaskError)
        )

    with pytest.raises(Exception) as e:
        print(pid_fn.get(pid_ref2))
        assert isinstance(
            e, (ray.exceptions.TaskCancelledError, ray.exceptions.RayTaskError)
        )


@pytest.mark.clustertest
def test_function_queueing(cpu_cluster):
    pid_fn = rh.function(getpid).to(cpu_cluster)

    pids = [pid_fn.enqueue(resources={"num_cpus": 1}) for _ in range(10)]
    assert len(pids) == 10


@pytest.mark.clustertest
def test_function_to_env(cpu_cluster):
    cpu_cluster.run(["pip uninstall numpy -y"])

    np_func = rh.function(np_array).to(cpu_cluster, env=["numpy"])

    list = [1, 2, 3]
    res = np_func(list)
    assert res.tolist() == list


@unittest.skip("Not working properly.")
@pytest.mark.clustertest
def test_function_external_fn(cpu_cluster):
    """Test functioning a module from reqs, not from working_dir"""
    import torch

    re_fn = rh.function(torch.sum).to(cpu_cluster, env=["torch"])
    res = re_fn(torch.arange(5))
    assert int(res) == 10


@unittest.skip("Runs indefinitely.")
@pytest.mark.clustertest
def test_notebook(cpu_cluster):
    nb_sum = lambda x: multiproc_torch_sum(x)
    re_fn = rh.function(nb_sum).to(cpu_cluster, env=["torch==1.12.1"])

    re_fn.notebook()
    summands = list(zip(range(5), range(4, 9)))
    res = re_fn(summands)

    assert res == [4, 6, 8, 10, 12]
    re_fn.delete_configs()


@unittest.skip("Runs indefinitely.")
def test_ssh():
    # TODO do this properly
    my_function = rh.function(name="local_function")
    my_function.ssh()
    assert True


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_share_function(cpu_cluster):
    my_function = rh.function(fn=summer).to(cpu_cluster).save(REMOTE_FUNC_NAME)

    my_function.share(
        users=["donny@run.house", "josh@run.house"],
        access_type="read",
        notify_users=False,
    )
    assert True


@pytest.mark.rnstest
def test_load_shared_function():
    my_function = rh.function(name=REMOTE_FUNC_NAME)
    res = my_function(1, 2)
    assert res == 3


@pytest.mark.rnstest
def delete_function_from_rns(s):
    server_url = s.rns_client.api_server_url
    resource_request_uri = s.rns_client.resource_uri(s.name)
    resp = requests.delete(
        f"{server_url}/resource/{resource_request_uri}",
        headers=s.rns_client.request_headers,
    )
    if resp.status_code != 200:
        raise Exception(
            f"Failed to delete_configs function data from path: {load_resp_content(resp)}"
        )

    try:
        # Terminate the cluster
        s.cluster.teardown_and_delete()
    except Exception as e:
        raise Exception(f"Failed to teardown the cluster: {e}")


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_byo_cluster_function(byo_cpu):
    re_fn = rh.function(multiproc_torch_sum).to(
        byo_cpu, env=["torch==1.12.1", "pytest"]
    )

    summands = list(zip(range(5), range(4, 9)))
    res = re_fn(summands)

    assert res == [4, 6, 8, 10, 12]


@pytest.mark.clustertest
def test_byo_cluster_maps(byo_cpu):
    pid_fn = rh.function(getpid).to(byo_cpu)
    num_pids = [1] * 20
    pids = pid_fn.map(num_pids)
    assert len(set(pids)) > 1
    assert all(pid > 0 for pid in pids)

    pid_run = pid_fn.run()

    pids = pid_fn.repeat(num_repeats=20)
    assert len(set(pids)) > 1
    assert all(pid > 0 for pid in pids)

    pids = [pid_fn.enqueue() for _ in range(10)]
    assert len(pids) == 10

    pid_res = pid_fn.get(pid_run.name)
    assert pid_res > 0

    # Test passing an objectref into a normal call
    pid_res_from_ref = pid_fn(pid_run.name)
    assert pid_res_from_ref > pid_res

    re_fn = rh.function(summer).to(byo_cpu)
    summands = list(zip(range(5), range(4, 9)))
    res = re_fn.starmap(summands)
    assert res == [4, 6, 8, 10, 12]


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_load_function_in_new_env(cpu_cluster, byo_cpu):
    cpu_cluster.save(
        f"@/{cpu_cluster.name}"
    )  # Needs to be saved to rns, right now has a local name by default
    remote_sum = rh.function(summer).to(cpu_cluster).save(REMOTE_FUNC_NAME)

    remote_python = (
        "import runhouse as rh; "
        f"remote_sum = rh.function(name='{REMOTE_FUNC_NAME}'); "
        "res = remote_sum(1, 5); "
        "assert res == 6"
    )
    res = byo_cpu.run_python([remote_python], stream_logs=True)
    assert res[0][0] == 0

    remote_sum.delete_configs()


@pytest.mark.clustertest
def test_nested_diff_clusters(cpu_cluster, byo_cpu):
    summer_cpu = rh.function(summer).to(cpu_cluster)
    call_function_diff_cpu = rh.function(call_function).to(byo_cpu)

    kwargs = {"a": 1, "b": 5}
    res = call_function_diff_cpu(summer_cpu, **kwargs)
    assert res == 6


@pytest.mark.clustertest
def test_nested_same_cluster(cpu_cluster):
    # When the system of a function is set to the cluster that the function is being called on, we run the function
    # locally and not via an RPC call
    summer_cpu = rh.function(fn=summer).to(system=cpu_cluster)
    call_function_cpu = rh.function(fn=call_function).to(system=cpu_cluster)

    kwargs = {"a": 1, "b": 5}
    res = call_function_cpu(summer_cpu, **kwargs)
    assert res == 6


@pytest.mark.clustertest
def test_http_url(cpu_cluster):
    rh.function(summer).to(cpu_cluster).save("@/remote_function")
    tun, port = cpu_cluster.ssh_tunnel(80, 50052)
    sum1 = requests.post(
        "http://127.0.0.1:80/call/remote_function/", json={"args": [1, 2]}
    ).json()
    assert sum1 == 3
    sum2 = requests.post(
        "http://127.0.0.1:80/call/remote_function/", json={"kwargs": {"a": 1, "b": 2}}
    ).json()
    assert sum2 == 3

    tun.close()


@unittest.skip("Not yet implemented.")
def test_http_url_with_curl():
    # NOTE: Assumes the Function has already been created and deployed to running cluster
    s = rh.function(name="test_function")
    curl_cmd = s.http_url(a=1, b=2, curl_command=True)
    print(curl_cmd)

    # delete_configs the function data from the RNS
    delete_function_from_rns(s)

    assert True


# test that deprecated arguments are still backwards compatible for now
@pytest.mark.clustertest
def test_reqs_backwards_compatible(cpu_cluster):
    summer_cpu = rh.function(fn=summer).to(system=cpu_cluster)
    res = summer_cpu(1, 5)
    assert res == 6

    torch_summer_cpu = rh.function(fn=summer).to(system=cpu_cluster, env=["torch"])
    torch_res = torch_summer_cpu(1, 5)
    assert torch_res == 6


@pytest.mark.clustertest
def test_setup_cmds_backwards_compatible(cpu_cluster):
    torch_summer_cpu = rh.function(fn=summer).to(system=cpu_cluster, env=["torch"])
    torch_res = torch_summer_cpu(1, 5)
    assert torch_res == 6


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: unary.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='unary.proto',
  package='unary',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0bunary.proto\x12\x05unary\"B\n\x07Message\x12\x0f\n\x07message\x18\x01 \x01(\x0c\x12\x13\n\x0bmodule_name\x18\x02 \x01(\t\x12\x11\n\tfunc_name\x18\x03 \x01(\t\"J\n\x12RunMessageResponse\x12\x0e\n\x06result\x18\x01 \x01(\x0c\x12\x11\n\texception\x18\x02 \x01(\x0c\x12\x11\n\ttraceback\x18\x03 \x01(\t\"I\n\x0fMessageResponse\x12\x0f\n\x07message\x18\x01 \x01(\x0c\x12\x10\n\x08received\x18\x02 \x01(\x08\x12\x13\n\x0boutput_type\x18\x03 \x01(\t2\xbd\x04\n\x05Unary\x12\x38\n\tRunModule\x12\x0e.unary.Message\x1a\x19.unary.RunMessageResponse\"\x00\x12;\n\x0fInstallPackages\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse\"\x00\x12\x35\n\tClearPins\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse\"\x00\x12\x35\n\tCancelRun\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse\"\x00\x12\x34\n\x08ListKeys\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse\"\x00\x12\x35\n\tPutObject\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse\"\x00\x12\x36\n\nAddSecrets\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse\"\x00\x12\x38\n\x0cGetRunObject\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse\"\x00\x12\x37\n\x0bRunCommands\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse\"\x00\x12\x37\n\tGetObject\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse\"\x00\x30\x01\x62\x06proto3'
)




_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='unary.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='unary.Message.message', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='module_name', full_name='unary.Message.module_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='func_name', full_name='unary.Message.func_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=88,
)


_RUNMESSAGERESPONSE = _descriptor.Descriptor(
  name='RunMessageResponse',
  full_name='unary.RunMessageResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='unary.RunMessageResponse.result', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exception', full_name='unary.RunMessageResponse.exception', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='traceback', full_name='unary.RunMessageResponse.traceback', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=90,
  serialized_end=164,
)


_MESSAGERESPONSE = _descriptor.Descriptor(
  name='MessageResponse',
  full_name='unary.MessageResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='unary.MessageResponse.message', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='received', full_name='unary.MessageResponse.received', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='output_type', full_name='unary.MessageResponse.output_type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=166,
  serialized_end=239,
)

DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
DESCRIPTOR.message_types_by_name['RunMessageResponse'] = _RUNMESSAGERESPONSE
DESCRIPTOR.message_types_by_name['MessageResponse'] = _MESSAGERESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'unary_pb2'
  # @@protoc_insertion_point(class_scope:unary.Message)
  })
_sym_db.RegisterMessage(Message)

RunMessageResponse = _reflection.GeneratedProtocolMessageType('RunMessageResponse', (_message.Message,), {
  'DESCRIPTOR' : _RUNMESSAGERESPONSE,
  '__module__' : 'unary_pb2'
  # @@protoc_insertion_point(class_scope:unary.RunMessageResponse)
  })
_sym_db.RegisterMessage(RunMessageResponse)

MessageResponse = _reflection.GeneratedProtocolMessageType('MessageResponse', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGERESPONSE,
  '__module__' : 'unary_pb2'
  # @@protoc_insertion_point(class_scope:unary.MessageResponse)
  })
_sym_db.RegisterMessage(MessageResponse)



_UNARY = _descriptor.ServiceDescriptor(
  name='Unary',
  full_name='unary.Unary',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=242,
  serialized_end=815,
  methods=[
  _descriptor.MethodDescriptor(
    name='RunModule',
    full_name='unary.Unary.RunModule',
    index=0,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_RUNMESSAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='InstallPackages',
    full_name='unary.Unary.InstallPackages',
    index=1,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_MESSAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ClearPins',
    full_name='unary.Unary.ClearPins',
    index=2,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_MESSAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CancelRun',
    full_name='unary.Unary.CancelRun',
    index=3,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_MESSAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListKeys',
    full_name='unary.Unary.ListKeys',
    index=4,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_MESSAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='PutObject',
    full_name='unary.Unary.PutObject',
    index=5,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_MESSAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='AddSecrets',
    full_name='unary.Unary.AddSecrets',
    index=6,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_MESSAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetRunObject',
    full_name='unary.Unary.GetRunObject',
    index=7,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_MESSAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RunCommands',
    full_name='unary.Unary.RunCommands',
    index=8,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_MESSAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetObject',
    full_name='unary.Unary.GetObject',
    index=9,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_MESSAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_UNARY)

DESCRIPTOR.services_by_name['Unary'] = _UNARY

# @@protoc_insertion_point(module_scope)



import subprocess
import webbrowser
from typing import Optional

import pkg_resources
import typer
from rich.console import Console

import runhouse.rns.login

from runhouse import cluster, configs

# create an explicit Typer application
app = typer.Typer(add_completion=False)
state = {"verbose": False}

# For printing with typer
console = Console()


@app.command()
def login(
    token: Optional[str] = typer.Argument(None, help="Your Runhouse API token"),
    yes: Optional[bool] = typer.Option(
        False, "--yes", "-y", help="Sets any confirmations to 'yes' automatically."
    ),
):
    """Login to Runhouse. Validates token provided, with options to upload or download stored secrets or config between
    local environment and Runhouse / Vault.
    """
    valid_token: str = (
        runhouse.rns.login.login(
            token=token,
            download_config=True,
            upload_config=True,
            download_secrets=True,
            upload_secrets=True,
        )
        if yes
        else runhouse.rns.login.login(token=token, interactive=True, ret_token=True)
    )

    if valid_token:
        webbrowser.open(f"{configs.get('dashboard_url')}/dashboard?token={valid_token}")
        raise typer.Exit()
    else:
        raise typer.Exit(code=1)


@app.command()
def logout():
    """Logout of Runhouse. Provides options to delete locally configured secrets and local Runhouse configs"""
    runhouse.rns.login.logout(interactive=True)
    raise typer.Exit()


@app.command()
def notebook(
    cluster_name: str, up: bool = typer.Option(False, help="Start the cluster")
):
    """Open a Jupyter notebook on a cluster."""
    c = cluster(name=cluster_name)
    if up:
        c.up_if_not()
    if not c.is_up():
        console.print(
            f"Cluster {cluster_name} is not up. Please run `runhouse notebook {cluster_name} --up`."
        )
        raise typer.Exit(1)
    c.notebook()


@app.command()
def ssh(cluster_name: str, up: bool = typer.Option(False, help="Start the cluster")):
    """SSH into a cluster created elsewhere (so `ssh cluster` doesn't work out of the box) or not yet up."""
    c = cluster(name=cluster_name)
    if up:
        c.up_if_not()
    if not c.is_up():
        console.print(
            f"Cluster {cluster_name} is not up. Please run `runhouse ssh {cluster_name} --up`."
        )
        raise typer.Exit(1)
    subprocess.call(f"ssh {c.name}", shell=True)


@app.command()
def cancel(
    cluster_name: str,
    run_key: str,
    force: Optional[bool] = typer.Option(False, help="Force cancel"),
    all: Optional[bool] = typer.Option(False, help="Cancel all jobs"),
):
    """Cancel a run on a cluster."""
    c = cluster(name=cluster_name)
    if all:
        c.cancel_all(force=force)
    else:
        c.cancel(run_key, force=force)


@app.command()
def logs(
    cluster_name: str,
    run_key: str,
    print_results: Optional[bool] = typer.Option(False, help="Print results"),
):
    """Get logs from a run on a cluster."""
    c = cluster(name=cluster_name)
    res = c.get(run_key, stream_logs=True)
    if print_results:
        console.print(res)


def load_cluster(cluster_name: str):
    """Load a cluster from RNS into the local environment, e.g. to be able to ssh."""
    c = cluster(name=cluster_name)
    if not c.address:
        c._update_from_sky_status(dryrun=True)


@app.command()
def start(
    restart_ray: bool = typer.Option(False, help="Restart the Ray runtime"),
    screen: bool = typer.Option(False, help="Start the server in a screen"),
):
    http_server_cmd = "python -m runhouse.servers.http.http_server"
    kill_proc_cmd = ["pkill", "-f", f"{http_server_cmd}"]
    subprocess.run(kill_proc_cmd)

    if restart_ray:
        subprocess.run(["ray", "stop"])
        subprocess.run(["ray", "start", "--head"])

    start_server_cmd = http_server_cmd.split()
    if screen:
        start_server_cmd = ["screen", "-dm", "bash", "-c"] + start_server_cmd
    subprocess.run(start_server_cmd)


@app.command()
def restart_server(
    cluster_name: str,
    restart_ray: bool = typer.Option(False, help="Restart the Ray runtime"),
    resync_rh: bool = typer.Option(False, help="Resync the Runhouse package"),
):
    """Restart the gRPC server on a cluster."""
    c = cluster(name=cluster_name)
    c.restart_server(resync_rh=resync_rh, restart_ray=restart_ray)


@app.callback()
def main(verbose: bool = False):
    """
    Runhouse CLI
    """
    if verbose:
        name = "runhouse"
        version = pkg_resources.get_distribution(name).version
        console.print(f"{name}=={version}", style="bold green")
        state["verbose"] = True



import shutil
import unittest

import pandas as pd
import pyarrow as pa
import pytest
import ray.data
import runhouse as rh

from runhouse import Folder

NUM_PARTITIONS = 10


def delete_local_folder(path):
    shutil.rmtree(path)


def tokenize_function(examples):
    from transformers import AutoTokenizer

    tokenizer = AutoTokenizer.from_pretrained("bert-base-cased")
    return tokenizer(examples["text"], padding="max_length", truncation=True)


# -----------------------------------------------
# ----------------- Local tests -----------------
# -----------------------------------------------
@pytest.mark.rnstest
def test_create_and_reload_file_locally(tmp_path):
    orig_data = pd.DataFrame({"my_col": list(range(50))})
    name = "~/my_local_test_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=str(tmp_path),
            system="file",
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data

    assert reloaded_data.to_pandas().equals(orig_data)

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pd.DataFrame)
        assert batch["my_col"].tolist() == list(range(idx * 10, (idx + 1) * 10))

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.rnstest
def test_create_and_reload_pandas_locally(pandas_table, tmp_path):
    name = "~/my_test_local_pandas_table"

    my_table = (
        rh.table(
            data=pandas_table,
            path=str(tmp_path),
            name=name,
            system="file",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data

    assert pandas_table.equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pd.DataFrame)
        assert batch["id"].tolist() == list(range(1, 7))

    del pandas_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.rnstest
def test_create_and_reload_pyarrow_locally(arrow_table, tmp_path):
    name = "~/my_test_local_pyarrow_table"

    my_table = (
        rh.table(
            data=arrow_table,
            name=name,
            path=str(tmp_path),
            system="file",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data

    assert arrow_table.to_pandas().equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pa.Table)
        assert batch["int"].to_pylist() == list(range(1, 11))

    del arrow_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.rnstest
def test_create_and_reload_ray_locally(ray_table, tmp_path):
    name = "~/my_test_local_ray_table"

    my_table = (
        rh.table(
            data=ray_table,
            path=str(tmp_path),
            name=name,
            system="file",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data

    assert ray_table.to_pandas().equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pa.Table)
        # NOTE [DG] 2021-08-10: This will generally fail because ray automatically partitions the data into
        # blocks, and order is not necessarily preserved when reading the data back in. Ideally we fix this
        # when we switch to in-memory tables.
        # assert batch["value"].to_pylist() == list(range(idx * 10, (idx + 1) * 10))

        if idx in [0, 10, 33]:
            # Some random batches to check
            assert [isinstance(val, int) for val in batch["value"].to_pylist()]

    del ray_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.rnstest
def test_create_and_reload_huggingface_locally(huggingface_table, tmp_path):
    name = "~/my_test_local_huggingface_table"

    my_table = (
        rh.table(
            data=huggingface_table,
            name=name,
            path=str(tmp_path),
            system="file",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data

    assert huggingface_table.to_pandas().equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10, as_dict=False)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["label", "text"]
        assert batch.shape == (10, 2)

    del huggingface_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.rnstest
def test_create_and_reload_dask_locally(dask_table, tmp_path):
    name = "~/my_test_local_dask_table"

    my_table = (
        rh.table(
            data=dask_table,
            name=name,
            path=str(tmp_path),
            system="file",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: "dask.dataframe.core.DataFrame" = reloaded_table.data.to_dask()
    assert reloaded_data.columns.to_list() == ["a", "b"]

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["a", "b"]
        assert batch.shape == (10, 2)

    del dask_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


# --------------------------------------------
# ----------------- S3 tests -----------------
# --------------------------------------------
@pytest.mark.awstest
@pytest.mark.rnstest
def test_create_and_reload_pyarrow_data_from_s3(arrow_table, table_s3_bucket):
    name = "@/my_test_pyarrow_table"

    my_table = (
        rh.table(
            data=arrow_table,
            name=name,
            path=f"/{table_s3_bucket}/pyarrow_df",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert reloaded_data.to_pandas().equals(arrow_table.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["int", "str"]

    del arrow_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.awstest
@pytest.mark.rnstest
def test_create_and_reload_pandas_data_from_s3(pandas_table, table_s3_bucket):
    name = "@/my_test_pandas_table"

    my_table = (
        rh.table(
            data=pandas_table,
            name=name,
            path=f"/{table_s3_bucket}/pandas_df",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert pandas_table.equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pd.DataFrame)
        assert batch["id"].tolist() == list(range(1, 7))

    del pandas_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.awstest
@pytest.mark.rnstest
def test_create_and_reload_huggingface_data_from_s3(huggingface_table, table_s3_bucket):
    name = "@/my_test_hf_table"

    my_table = (
        rh.table(
            data=huggingface_table,
            name=name,
            path=f"/{table_s3_bucket}/huggingface_data",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)

    # Stream in as huggingface dataset
    batches = reloaded_table.stream(batch_size=10, as_dict=False)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["label", "text"]
        assert batch.shape == (10, 2)

    del huggingface_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.awstest
@pytest.mark.rnstest
def test_create_and_reload_dask_data_from_s3(dask_table, table_s3_bucket):
    name = "@/my_test_dask_table"

    my_table = (
        rh.table(
            data=dask_table,
            name=name,
            path=f"/{table_s3_bucket}/dask",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: "dask.dataframe.core.DataFrame" = reloaded_table.data.to_dask()
    assert reloaded_data.columns.to_list() == ["a", "b"]

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["a", "b"]
        assert batch.shape == (10, 2)

    del dask_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.awstest
@pytest.mark.rnstest
def test_create_and_reload_ray_data_from_s3(ray_table, table_s3_bucket):
    name = "@/my_test_ray_table"

    my_table = (
        rh.table(
            data=ray_table,
            name=name,
            path=f"/{table_s3_bucket}/ray_data",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert reloaded_data.to_pandas().equals(ray_table.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pa.Table)
        if idx in [0, 10, 33]:
            # Some random batches to check
            assert [isinstance(val, int) for val in batch["value"].to_pylist()]

    del ray_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


# ----------------- Iter -----------------
@pytest.mark.awstest
@pytest.mark.rnstest
def test_load_pandas_data_as_iter(pandas_table, table_s3_bucket):
    name = "@/my_test_pandas_table"

    my_table = (
        rh.table(
            data=pandas_table,
            name=name,
            path=f"/{table_s3_bucket}/pandas",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = next(iter(reloaded_table))

    assert isinstance(reloaded_data, pd.Series)
    assert reloaded_data.to_dict() == {"id": 1, "grade": "a"}

    del pandas_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.awstest
@pytest.mark.rnstest
def test_load_pyarrow_data_as_iter(arrow_table, table_s3_bucket):
    name = "@/my_test_pyarrow_table"

    my_table = (
        rh.table(
            data=arrow_table,
            name=name,
            path=f"/{table_s3_bucket}/pyarrow-data",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: pa.ChunkedArray = next(iter(reloaded_table))

    assert isinstance(reloaded_data, pa.ChunkedArray)
    assert reloaded_data.to_pylist() == list(range(1, 11))

    del arrow_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.awstest
@pytest.mark.rnstest
def test_load_huggingface_data_as_iter(huggingface_table, table_s3_bucket):
    name = "@/my_test_huggingface_table"

    my_table = (
        rh.table(
            data=huggingface_table,
            name=name,
            path=f"/{table_s3_bucket}/huggingface-dataset",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: pa.ChunkedArray = next(iter(reloaded_table))
    assert isinstance(reloaded_data, pa.ChunkedArray)

    del huggingface_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


# ----------------- Shuffling -----------------
@pytest.mark.awstest
@pytest.mark.rnstest
def test_shuffling_pyarrow_data_from_s3(arrow_table, table_s3_bucket):
    name = "@/my_test_shuffled_pyarrow_table"

    my_table = (
        rh.table(
            data=arrow_table,
            name=name,
            path=f"/{table_s3_bucket}/pyarrow",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    batches = reloaded_table.stream(
        batch_size=10, shuffle_seed=42, shuffle_buffer_size=10
    )
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pa.Table)
        assert arrow_table.columns[0].to_pylist() != batch.columns[0].to_pylist()
        assert arrow_table.columns[1].to_pylist() != batch.columns[1].to_pylist()

    del arrow_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


# -------------------------------------------------
# ----------------- Cluster tests -----------------
# -------------------------------------------------
@pytest.mark.clustertest
@pytest.mark.rnstest
def test_create_and_reload_pandas_data_from_cluster(pandas_table, cpu_cluster):
    # Make sure the destination folder for the data exists on the cluster
    data_path_on_cluster = f"{Folder.DEFAULT_CACHE_FOLDER}/pandas-data"
    cpu_cluster.run([f"mkdir -p {data_path_on_cluster}"])

    name = "@/my_test_pandas_table"
    my_table = (
        rh.table(
            data=pandas_table,
            name=name,
            path=data_path_on_cluster,
            system=cpu_cluster,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)

    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert pandas_table.equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pd.DataFrame)
        assert batch["id"].tolist() == list(range(1, 7))

    del pandas_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_create_and_reload_ray_data_from_cluster(ray_table, cpu_cluster):
    data_path_on_cluster = f"{Folder.DEFAULT_CACHE_FOLDER}/ray-data"
    cpu_cluster.run([f"mkdir -p {data_path_on_cluster}"])

    name = "@/my_test_ray_cluster_table"

    my_table = (
        rh.table(
            data=ray_table,
            name=name,
            path=data_path_on_cluster,
            system=cpu_cluster,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert ray_table.to_pandas().equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pa.Table)
        if idx in [0, 10, 33]:
            # Some random batches to check
            assert [isinstance(val, int) for val in batch["value"].to_pylist()]

    del ray_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_create_and_reload_pyarrow_data_from_cluster(arrow_table, cpu_cluster):
    data_path_on_cluster = f"{Folder.DEFAULT_CACHE_FOLDER}/pyarrow-data"
    cpu_cluster.run([f"mkdir -p {data_path_on_cluster}"])

    name = "@/my_test_pyarrow_cluster_table"

    my_table = (
        rh.table(
            data=arrow_table,
            name=name,
            path=data_path_on_cluster,
            system=cpu_cluster,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert arrow_table.to_pandas().equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["int", "str"]

    del arrow_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_create_and_reload_huggingface_data_from_cluster(
    huggingface_table, cpu_cluster
):
    data_path_on_cluster = f"{Folder.DEFAULT_CACHE_FOLDER}/hf-data"
    cpu_cluster.run([f"mkdir -p {data_path_on_cluster}"])

    name = "@/my_test_hf_cluster_table"

    my_table = (
        rh.table(
            data=huggingface_table,
            name=name,
            path=data_path_on_cluster,
            system=cpu_cluster,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert huggingface_table.to_pandas().equals(reloaded_data.to_pandas())

    # Stream in as huggingface dataset
    batches = reloaded_table.stream(batch_size=10, as_dict=False)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["label", "text"]
        assert batch.shape == (10, 2)

    del huggingface_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_create_and_reload_dask_data_from_cluster(dask_table, cpu_cluster):
    data_path_on_cluster = f"{Folder.DEFAULT_CACHE_FOLDER}/dask-data"
    cpu_cluster.run([f"mkdir -p {data_path_on_cluster}"])

    name = "@/my_test_dask_cluster_table"

    my_table = (
        rh.table(
            data=dask_table,
            name=name,
            path=data_path_on_cluster,
            system=cpu_cluster,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: "dask.dataframe.core.DataFrame" = reloaded_table.data.to_dask()
    assert reloaded_data.columns.to_list() == ["a", "b"]

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["a", "b"]
        assert batch.shape == (10, 2)

    del dask_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.clustertest
@pytest.mark.rnstest
def test_to_cluster_attr(pandas_table, cpu_cluster, tmp_path):
    local_path = tmp_path / "table_tests/local_test_table"
    name = "~/my_local_test_table"

    my_table = (
        rh.table(
            data=pandas_table,
            name=name,
            path=str(local_path),
            system="file",
        )
        .write()
        .save()
    )

    cluster_table = my_table.to(system=cpu_cluster)

    assert isinstance(cluster_table.system, rh.Cluster)
    assert cluster_table._folder._fs_str == "ssh"

    data = cluster_table.data
    assert data.to_pandas().equals(pandas_table)

    del pandas_table
    del my_table

    cluster_table.delete_configs()
    cluster_table.rm()
    assert not cluster_table.exists_in_system()


# -------------------------------------------------
# ----------------- Fetching tests -----------------
# -------------------------------------------------
@pytest.mark.awstest
@pytest.mark.rnstest
def test_create_and_fetch_pyarrow_data_from_s3(arrow_table, table_s3_bucket):
    name = "@/my_test_fetch_pyarrow_table"

    my_table = (
        rh.table(
            data=arrow_table,
            name=name,
            path=f"/{table_s3_bucket}/pyarrow",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: pa.Table = reloaded_table.fetch()
    assert arrow_table == reloaded_data

    del arrow_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.awstest
@pytest.mark.rnstest
def test_create_and_fetch_pandas_data_from_s3(pandas_table, table_s3_bucket):
    name = "@/my_test_fetch_pandas_table"

    my_table = (
        rh.table(
            data=pandas_table,
            name=name,
            path=f"/{table_s3_bucket}/pandas",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: pd.DataFrame = reloaded_table.fetch()
    assert pandas_table.equals(reloaded_data)

    del pandas_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.awstest
@pytest.mark.rnstest
def test_create_and_fetch_huggingface_data_from_s3(huggingface_table, table_s3_bucket):
    name = "@/my_test_fetch_huggingface_table"

    my_table = (
        rh.table(
            data=huggingface_table,
            name=name,
            path=f"/{table_s3_bucket}/huggingface",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data = reloaded_table.fetch()
    assert huggingface_table.shape == reloaded_data.shape

    del huggingface_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.awstest
@pytest.mark.rnstest
def test_create_and_fetch_ray_data_from_s3(ray_table, table_s3_bucket):
    name = "@/my_test_fetch_ray_table"

    my_table = (
        rh.table(
            data=ray_table,
            name=name,
            path=f"/{table_s3_bucket}/ray",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: ray.data.Dataset = reloaded_table.fetch()
    assert ray_table.to_pandas().equals(reloaded_data.to_pandas())

    del ray_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


@pytest.mark.awstest
@pytest.mark.rnstest
def test_create_and_fetch_dask_data_from_s3(dask_table, table_s3_bucket):
    name = "@/my_test_fetch_dask_table"

    my_table = (
        rh.table(
            data=dask_table,
            name=name,
            path=f"/{table_s3_bucket}/dask",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.table(name=name)
    reloaded_data: "dask.dataframe.core.DataFrame" = reloaded_table.fetch()
    assert dask_table.npartitions == reloaded_data.npartitions

    del dask_table
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.rm()
    assert not reloaded_table.exists_in_system()


# -------------------------------------------------
# ----------------- Table Sharing tests -----------------
# -------------------------------------------------
@pytest.mark.rnstest
def test_sharing_table(pandas_table):
    name = "shared_pandas_table"

    my_table = (
        rh.table(
            data=pandas_table,
            name=name,
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    my_table.share(
        users=["donny@run.house", "josh@run.house"],
        access_type="write",
        notify_users=False,
    )

    assert my_table.exists_in_system()


@pytest.mark.rnstest
def test_read_shared_table():
    my_table = rh.table(name="@/shared_pandas_table")
    df: pd.DataFrame = my_table.fetch()
    assert not df.empty


if __name__ == "__main__":
    unittest.main()



import copy
import logging
import re
import subprocess
import sys
from pathlib import Path
from typing import Dict, Optional, Union

from runhouse import rh_config
from runhouse.rns.folders import Folder, folder
from runhouse.rns.resource import Resource
from runhouse.rns.utils.hardware import _get_cluster_from

INSTALL_METHODS = {"local", "reqs", "pip", "conda"}

logger = logging.getLogger(__name__)


class Package(Resource):
    RESOURCE_TYPE = "package"

    # https://pytorch.org/get-started/locally/
    # Note: no binaries exist for 11.4 (https://github.com/pytorch/pytorch/issues/75992)
    TORCH_INDEX_URLS = {
        "11.3": "https://download.pytorch.org/whl/cu113",
        "11.5": "https://download.pytorch.org/whl/cu115",
        "11.6": "https://download.pytorch.org/whl/cu116",
        "11.7": "https://download.pytorch.org/whl/cu117",
        "11.8": "https://download.pytorch.org/whl/cu118",
        "cpu": "https://download.pytorch.org/whl/cpu",
    }

    def __init__(
        self,
        name: str = None,
        install_method: str = None,
        install_target: Union[str, Folder] = None,
        install_args: str = None,
        dryrun: bool = False,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse Package resource.
        """
        super().__init__(
            name=name,
            dryrun=dryrun,
        )
        self.install_method = install_method
        self.install_target = install_target
        self.install_args = install_args

    @property
    def config_for_rns(self):
        # If the package is just a simple Package.from_string string, no
        # need to store it in rns, just give back the string.
        # if self.install_method in ['pip', 'conda', 'git']:
        #     return f'{self.install_method}:{self.name}'
        config = super().config_for_rns
        config["install_method"] = self.install_method
        config["install_target"] = self._resource_string_for_subconfig(
            self.install_target
        )
        config["install_args"] = self.install_args
        return config

    def __str__(self):
        if self.name:
            return f"Package: {self.name}"
        if isinstance(self.install_target, Folder):
            # if self.install_target.name:
            #     return f'Package: {self.install_target.name}'
            return f"Package: {self.install_target.path}"
        return f"Package: {self.install_target}"

    def _install(self, env: Union[str, "Env"] = None):
        """Install package.

        Args:
            env (Env or str): Environment to install package on. If left empty, defaults to base environment.
                (Default: ``None``)
        """
        logging.info(
            f"Installing package {str(self)} with method {self.install_method}."
        )
        install_cmd = ""
        install_args = f" {self.install_args}" if self.install_args else ""
        cuda_version_or_cpu = self._detect_cuda_version_or_cpu()

        if isinstance(self.install_target, Folder):
            local_path = self.install_target.local_path
            if self.install_method == "pip":
                # TODO [DG] Revisit: Would be nice if we could use -e by default, but importlib on the rpc server
                #  isn't finding the package right after its installed.
                # if (Path(local_path) / 'setup.py').exists():
                #     install_cmd = f'-e {local_path}' + install_args
                # else:
                install_cmd = f"{local_path}" + install_args
            elif self.install_method == "conda":
                install_cmd = f"{local_path}" + install_args
            elif self.install_method == "reqs":
                reqs_path = f"{local_path}/requirements.txt"
                logging.info(f"reqs path: {reqs_path}")
                if Path(reqs_path).expanduser().exists():
                    install_cmd = self._requirements_txt_install_cmd(
                        path=reqs_path,
                        cuda_version_or_cpu=cuda_version_or_cpu,
                        args=install_args,
                    )
                    logging.info(
                        f"pip installing requirements from {reqs_path} with: {install_cmd}"
                    )
                    self._pip_install(install_cmd)
                else:
                    logging.info(f"{local_path}/requirements.txt not found, skipping")
        else:
            install_cmd = self.install_target + install_args

        if self.install_method == "pip":
            install_cmd = self._install_cmd_for_torch(install_cmd, cuda_version_or_cpu)
            if not install_cmd:
                raise ValueError("Invalid install command")

            self._pip_install(install_cmd, env)
        elif self.install_method == "conda":
            self._conda_install(install_cmd, env)
        elif self.install_method in ["local", "reqs"]:
            if isinstance(self.install_target, Folder):
                sys.path.append(local_path)
            elif Path(self.install_target).resolve().expanduser().exists():
                sys.path.append(str(Path(self.install_target).resolve().expanduser()))
            else:
                raise ValueError(
                    f"install_target must be a Folder or a path to a directory for "
                    f"install_method {self.install_method}"
                )
        # elif self.install_method == 'unpickle':
        #     # unpickle the functions and make them importable
        #     with self.get('functions.pickle') as f:
        #         sys.modules[self.name] = pickle.load(f)
        else:
            raise ValueError(
                f"Unknown install_method {self.install_method}. Try using cluster.run() or to install instead."
            )

    # ----------------------------------
    # Torch Install Helpers
    # ----------------------------------
    def _requirements_txt_install_cmd(self, path, cuda_version_or_cpu="", args=""):
        """Read requirements from file, append --index-url and --extra-index-url where relevant for torch packages,
        and return list of formatted packages."""
        with open(path) as f:
            reqs = f.readlines()

        # if torch extra index url is already defined by the user or torch isn't a req, directly pip install reqs file
        if not [req for req in reqs if "torch" in req]:
            return f"-r {path}" + args
        for req in reqs:
            if (
                "--index-url" in req or "--extra-index-url" in req
            ) and "pytorch.org" in req:
                return f"-r {path}" + args

        # add extra-index-url for torch if not found
        return (
            f"-r {path} --extra-index-url {self._torch_index_url(cuda_version_or_cpu)}"
        )

    def _install_cmd_for_torch(self, install_cmd, cuda_version_or_cpu):
        """Return the correct formatted pip install command for the torch package(s) provided."""
        if install_cmd.startswith("#"):
            return None

        torch_source_packages = ["torch", "torchvision", "torchaudio"]
        if not any([x in install_cmd for x in torch_source_packages]):
            return install_cmd

        packages_to_install: list = self._packages_to_install_from_cmd(install_cmd)
        final_install_cmd = ""
        for package_install_cmd in packages_to_install:
            formatted_cmd = self._install_url_for_torch_package(
                package_install_cmd, cuda_version_or_cpu
            )
            if formatted_cmd:
                final_install_cmd += formatted_cmd + " "

        final_install_cmd = final_install_cmd.rstrip()
        return final_install_cmd if final_install_cmd != "" else None

    def _install_url_for_torch_package(self, install_cmd, cuda_version_or_cpu):
        """Build the full install command, adding a --index-url and --extra-index-url where applicable."""
        # Grab the relevant index url for torch based on the CUDA version provided
        if "," in install_cmd:
            # If installing a range of versions format the string to make it compatible with `pip_install` method
            install_cmd = install_cmd.replace(" ", "")

        index_url = self._torch_index_url(cuda_version_or_cpu)
        if index_url and not any(
            specifier in install_cmd for specifier in ["--index-url ", "-i "]
        ):
            install_cmd = f"{install_cmd} --index-url {index_url}"

        if "--extra-index-url" not in install_cmd:
            return f"{install_cmd} --extra-index-url https://pypi.python.org/simple/"

        return install_cmd

    def _torch_index_url(self, cuda_version_or_cpu: str):
        return self.TORCH_INDEX_URLS.get(cuda_version_or_cpu)

    @staticmethod
    def _detect_cuda_version_or_cpu():
        """Return the CUDA version on the cluster. If we are on a CPU-only cluster return 'cpu'.

        Note: A cpu-only machine may have the CUDA toolkit installed, which means nvcc will still return
        a valid version. Also check if the NVIDIA driver is installed to confirm we are on a GPU."""
        try:
            cuda_version_info: str = subprocess.check_output(
                "nvcc --version", shell=True
            ).decode("utf-8")
            cuda_version = cuda_version_info.split("release ")[1].split(",")[0]
        except subprocess.CalledProcessError:
            return "cpu"

        try:
            subprocess.check_call(["nvidia-smi"])
            return cuda_version
        except subprocess.CalledProcessError:
            return "cpu"

    @staticmethod
    def _packages_to_install_from_cmd(install_cmd: str):
        """Split a string of command(s) into a list of separate commands"""
        # Remove any --extra-index-url flags from the install command (to be added later by default)
        install_cmd = re.sub(r"--extra-index-url\s+\S+", "", install_cmd)
        install_cmd = install_cmd.strip()

        if ", " in install_cmd:
            # Ex: 'torch>=1.13.0,<2.0.0'
            return [install_cmd]

        matches = re.findall(r"(\S+(?:\s+(-i|--index-url)\s+\S+)?)", install_cmd)

        packages_to_install = [match[0] for match in matches]
        return packages_to_install

    # ----------------------------------

    @staticmethod
    def _pip_install(install_cmd: str, env: Union[str, "Env"] = ""):
        """Run pip install."""
        if env:
            if isinstance(env, str):
                from runhouse.rns.envs import Env

                env = Env.from_name(env)
            cmd_prefix = env._run_cmd
        else:
            cmd_prefix = f"{sys.executable} -m"
        cmd = f"{cmd_prefix} pip install {install_cmd}"
        logging.info(f"Running: {cmd}")
        subprocess.check_call(cmd.split(" "))

    @staticmethod
    def _conda_install(install_cmd: str, env: Union[str, "Env"] = ""):
        """Run conda install."""
        cmd = f"conda install -y {install_cmd}"
        if env:
            if isinstance(env, str):
                from runhouse.rns.envs import Env

                env = Env.from_name(env)
            cmd = f"{env._run_cmd} {cmd}"
        logging.info(f"Running: {cmd}")
        # check if conda is installed, and if not, install it
        try:
            subprocess.check_call(["conda", "--version"])
            subprocess.run(cmd.split(" "))
        except FileNotFoundError:
            logging.info("Conda not found, installing...")
            subprocess.check_call(
                "wget https://repo.continuum.io/miniconda/Miniconda3-latest-Linux-x86_64.sh "
                "-O ~/miniconda.sh".split(" ")
            )
            subprocess.check_call(["bash", "~/miniconda.sh", "-b", "-p", "~/miniconda"])
            subprocess.check_call("source $HOME/miniconda3/bin/activate".split(" "))
            status = subprocess.check_call(cmd.split(" "))
            if not status == 0:
                raise RuntimeError(
                    "Conda install failed, check that the package exists and is "
                    "available for your platform."
                )

    def to(
        self,
        system: Union[str, Dict, "Cluster"],
        path: Optional[str] = None,
        mount: bool = False,
    ):
        """Copy the package onto filesystem or cluster, and return the new Package object."""
        if not isinstance(self.install_target, Folder):
            raise TypeError(
                "`install_target` must be a Folder in order to copy the package to a system."
            )
        system = _get_cluster_from(system)
        if self.install_target.system == system:
            return self

        if isinstance(system, Resource):
            logger.info(
                f"Copying folder from {self.install_target.fsspec_url} to: {getattr(system, 'name', system)}"
            )
            new_folder = self.install_target._to_cluster(system, path=path, mount=mount)
        else:  # to fs
            new_folder = self.install_target.to(system, path=path)
        new_folder.system = system
        new_package = copy.copy(self)
        new_package.install_target = new_folder
        return new_package

    @staticmethod
    def from_config(config: dict, dryrun=False):
        if isinstance(config.get("install_target"), dict):
            config["install_target"] = Folder.from_config(
                config["install_target"], dryrun=dryrun
            )
        if config.get("resource_subtype") == "GitPackage":
            from runhouse import GitPackage

            return GitPackage.from_config(config, dryrun=dryrun)
        return Package(**config, dryrun=dryrun)

    @staticmethod
    def from_string(specifier: str, dryrun=False):
        if specifier == "requirements.txt":
            specifier = "reqs:./"

        # Use regex to check if specifier matches '<method>:https://github.com/<path>' or 'https://github.com/<path>'
        match = re.search(
            r"^(?:(?P<method>[^:]+):)?(?P<path>https://github.com/.+)", specifier
        )
        if match:
            install_method = match.group("method")
            url = match.group("path")
            from runhouse.rns.packages.git_package import git_package

            return git_package(
                git_url=url, install_method=install_method, dryrun=dryrun
            )

        target_and_args = specifier
        if specifier.split(":")[0] in INSTALL_METHODS:
            target_and_args = specifier.split(":", 1)[1]
        rel_target, args = (
            target_and_args.split(" ", 1)
            if " " in target_and_args
            else (target_and_args, "")
        )

        # We need to do this because relative paths are relative to the current working directory!
        abs_target = (
            Path(rel_target).expanduser()
            if Path(rel_target).expanduser().is_absolute()
            else Path(rh_config.rns_client.locate_working_dir()) / rel_target
        )
        if abs_target.exists():
            target = Folder(
                path=abs_target, dryrun=True
            )  # No need to create the folder here
        else:
            target = rel_target

        if specifier.startswith("local:"):
            return Package(install_target=target, install_method="local", dryrun=dryrun)
        elif specifier.startswith("reqs:"):
            return Package(
                install_target=target,
                install_args=args,
                install_method="reqs",
                dryrun=dryrun,
            )
        elif specifier.startswith("pip:"):
            return Package(
                install_target=target,
                install_args=args,
                install_method="pip",
                dryrun=dryrun,
            )
        elif specifier.startswith("conda:"):
            return Package(
                install_target=target,
                install_args=args,
                install_method="conda",
                dryrun=dryrun,
            )
        elif specifier.startswith("rh:"):
            # Calling the factory method below
            return package(name=specifier[3:], dryrun=dryrun)
        else:
            if Path(specifier).resolve().exists():
                return Package(
                    install_target=target,
                    install_args=args,
                    install_method="reqs",
                    dryrun=dryrun,
                )
            else:
                return Package(
                    install_target=target,
                    install_args=args,
                    install_method="pip",
                    dryrun=dryrun,
                )


def package(
    name: str = None,
    install_method: str = None,
    install_str: str = None,
    path: str = None,
    system: str = None,
    dryrun: bool = False,
    local_mount: bool = False,
    data_config: Optional[Dict] = None,
) -> Package:
    """
    Builds an instance of :class:`Package`.

    Args:
        name (str): Name to assign the package resource.
        install_method (str): Method for installing the package. Options: [``pip``, ``conda``, ``reqs``, ``local``]
        install_str (str): Additional arguments to install.
        path (str): URL of the package to install.
        system (str): File system or cluster on which the package lives. Currently this must a cluster or one of:
            [``file``, ``github``, ``sftp``, ``ssh``, ``s3``, ``gs``, ``azure``].
        dryrun (bool): Whether to create the Package if it doesn't exist, or load the Package object as a dryrun.
            (Default: ``False``)
        local_mount (bool): Whether to locally mount the installed package. (Default: ``False``)
        data_config (Optional[Dict]): The data config to pass to the underlying fsspec handler.

    Returns:
        Package: The resulting package.

    Example:
        >>> import runhouse as rh
        >>> reloaded_package = rh.package(name="my-package")
        >>> local_package = rh.package(path="local/folder/path", install_method="local")
    """
    if name and not any(
        [install_method, install_str, path, system, data_config, local_mount]
    ):
        # If only the name is provided and dryrun is set to True
        return Package.from_name(name, dryrun)

    install_target = None
    install_args = None
    if path is not None:
        system = system or Folder.DEFAULT_FS
        install_target = folder(
            path=path, system=system, local_mount=local_mount, data_config=data_config
        )
        install_args = install_str
    elif install_str is not None:
        install_target, install_args = install_str.split(" ", 1)

    return Package(
        install_method=install_method,
        install_target=install_target,
        install_args=install_args,
        name=name,
        dryrun=dryrun,
    )



import copy
import json
import logging
import sys
from datetime import datetime
from enum import Enum
from io import StringIO
from typing import Any, Optional, Union

from ray import cloudpickle as pickle

from runhouse import blob
from runhouse.rh_config import obj_store, rns_client
from runhouse.rns.api_utils.utils import log_timestamp, resolve_absolute_path

# Need to alias so it doesn't conflict with the folder property
from runhouse.rns.folders import Folder, folder as folder_factory
from runhouse.rns.hardware import Cluster
from runhouse.rns.resource import Resource
from runhouse.rns.top_level_rns_fns import resolve_rns_path
from runhouse.rns.utils.hardware import _current_cluster, _get_cluster_from

logger = logging.getLogger(__name__)


class RunStatus(str, Enum):
    NOT_STARTED = "NOT_STARTED"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    ERROR = "ERROR"


class RunType(str, Enum):
    CMD_RUN = "CMD"
    FUNCTION_RUN = "FUNCTION"
    CTX_MANAGER = "CTX_MANAGER"


class Run(Resource):
    RESOURCE_TYPE = "run"

    LOCAL_RUN_PATH = f"{rns_client.rh_directory}/runs"

    RUN_CONFIG_FILE = "config_for_run.json"
    RESULT_FILE = "result.pkl"
    INPUTS_FILE = "inputs.pkl"

    def __init__(
        self,
        name: str = None,
        fn_name: str = None,
        cmds: list = None,
        path: str = None,
        system: Union[str, Cluster] = None,
        data_config: dict = None,
        overwrite: bool = False,
        dryrun: bool = False,
        **kwargs,
    ):
        """
        Runhouse Run object

        .. note::
            To load an existing Run, please use the factory method :func:`run`.
        """
        run_name = name or str(self._current_timestamp())
        super().__init__(name=run_name, dryrun=dryrun)

        folder_system = system or Folder.DEFAULT_FS
        folder_path = (
            resolve_absolute_path(path)
            if path
            else (
                self._base_local_folder_path(self.name)
                if folder_system == Folder.DEFAULT_FS
                else self._base_cluster_folder_path(name=run_name)
            )
        )

        if overwrite:
            # Delete the Run from the system if one already exists
            self._delete_existing_run(folder_path, folder_system)

        # Create new folder which lives on the system and contains all the Run's data:
        # (run config, stdout, stderr, inputs, result)
        self.folder = folder_factory(
            path=folder_path,
            system=folder_system,
            data_config=data_config,
            dryrun=dryrun,
        )

        # Use the config for this run saved on the system if it exists
        run_config: dict = self._load_run_config(folder=self.folder)

        self.status = run_config.get("status", RunStatus.NOT_STARTED)
        self.start_time = run_config.get("start_time", None)
        self.end_time = run_config.get("end_time", None)
        self.upstream_artifacts = run_config.get("upstream_artifacts", [])
        self.downstream_artifacts = run_config.get("downstream_artifacts", [])
        self.fn_name = run_config.get("fn_name", fn_name)
        self.cmds = run_config.get("cmds", cmds)
        self.run_type = run_config.get("run_type", self._detect_run_type())

        self._stdout_path = self._path_to_file_by_ext(ext=".out")
        self._stderr_path = self._path_to_file_by_ext(ext=".err")

    def __enter__(self):
        self.status = RunStatus.RUNNING
        self.start_time = self._current_timestamp()

        # Begin tracking the Run in the rns_client - this adds the current Run to the stack of active Runs
        rns_client.start_run(self)

        sys.stdout = StringIO()
        sys.stderr = StringIO()

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.end_time = self._current_timestamp()
        self.status = RunStatus.COMPLETED

        # Pop the current Run from the stack of active Runs
        rns_client.stop_run()

        # Save Run config to its folder on the system - this will already happen on the cluster
        # for function based Runs
        self._write_config()

        if self.run_type in [RunType.FUNCTION_RUN, RunType.CMD_RUN]:
            # For function based Runs we use the logfiles already generated for the current Ray worker
            # on the cluster, and for cmd runs we are using the SSH command runner to get the stdout / stderr
            return

        stdout = sys.stdout.getvalue()
        stderr = sys.stderr.getvalue()

        # save stdout and stderr to their respective log files
        self.write(data=stdout.encode(), path=self._stdout_path)
        self.write(data=stderr.encode(), path=self._stderr_path)

        # return False to propagate any exception that occurred inside the with block
        return False

    @staticmethod
    def from_config(config: dict, dryrun=False):
        return Run(**config, dryrun=dryrun)

    @property
    def config_for_rns(self):
        """Metadata to store in RNS for the Run."""
        config = super().config_for_rns
        base_config = {
            "path": self.folder.path,
            "system": self._resource_string_for_subconfig(self.folder.system),
            "run_type": self.run_type,
        }
        config.update(base_config)
        return config

    @property
    def run_config(self):
        """Config to save in the Run's dedicated folder on the system.
        Note: this is different from the config saved in RNS, which is the metadata for the Run.
        """
        config = {
            "name": self.name,
            "status": self.status,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "run_type": self.run_type,
            "fn_name": self.fn_name,
            "cmds": self.cmds,
            # NOTE: artifacts are currently only tracked in context manager based runs
            "upstream_artifacts": self.upstream_artifacts,
            "downstream_artifacts": self.downstream_artifacts,
        }
        return config

    def save(
        self,
        name: str = None,
        overwrite: bool = True,
    ):
        """If the Run name is being overwritten (ex: initially created with auto-generated name),
        update the Run config stored on the system before saving to RNS."""
        config_for_rns = self.config_for_rns
        config_path = self._path_to_config()
        if not config_for_rns["name"] or name:
            config_for_rns["name"] = resolve_rns_path(name or self.name)
            self._write_config(config=config_for_rns)
            logger.info(f"Updated Run config name in path: {config_path}")

        return super().save(name, overwrite)

    def write(self, data: Any, path: str):
        """Write data (ex: function inputs or result, stdout, stderr) to the Run's dedicated folder on the system."""
        blob(data=data, system=self.folder.system, path=path).write()

    def to(
        self,
        system,
        path: Optional[str] = None,
        data_config: Optional[dict] = None,
    ):
        """Send a Run to another system.

        Args:
            system (Union[str or Cluster]): Name of the system or Cluster object to copy the Run to.
            path (Optional[str]): Path to the on the system to save the Run.
                Defaults to the local path for Runs (in the rh folder of the working directory).
            data_config (Optional[dict]): Config to pass into fsspec handler for copying the Run.

        Returns:
            Run: A copy of the Run on the destination system and path.
        """
        # TODO: [JL] - support for `on_completion` (wait to copy the results to destination until async run completes)

        new_run = copy.copy(self)

        if self.run_type == RunType.FUNCTION_RUN:
            results_path = self._fn_result_path()
            # Pickled function result should be saved down to the Run's folder on the cluster
            if results_path not in self.folder.ls():
                raise FileNotFoundError(
                    f"No results saved down in path: {results_path}"
                )

        for fp in [self._stdout_path, self._stderr_path]:
            # Stdout and Stderr files created on a cluster can be symlinks to the files that we create via Ray
            # by default - before copying them to a new system make sure they are regular files
            self._convert_symlink_to_file(path=fp)

        if system == "here":
            # Save to default local path if none provided
            path = path or self._base_local_folder_path(self.name)

        new_run.folder = self.folder.to(
            system=system, path=path, data_config=data_config
        )

        return new_run

    def refresh(self) -> "Run":
        """Reload the Run object from the system. This is useful for checking the status of a Run.
        For example: ``my_run.refresh().status``"""
        run_config = self._load_run_config(folder=self.folder)
        # Need the metadata from RNS and the Run specific data in order to re-load the Run object
        config = {**self.config_for_rns, **run_config}
        return Run.from_config(config, dryrun=True)

    def inputs(self) -> bytes:
        """Load the pickled function inputs saved on the system for the Run."""
        return pickle.loads(
            self._load_blob_from_path(path=self._fn_inputs_path()).fetch()
        )

    def result(self):
        """Load the function result saved on the system for the Run. If the Run has failed return the stderr,
        otherwise return the stdout."""
        run_status = self.refresh().status
        if run_status == RunStatus.COMPLETED:
            results_path = self._fn_result_path()
            if results_path not in self.folder.ls():
                raise FileNotFoundError(
                    f"No results file found in path: {results_path}"
                )
            return pickle.loads(self._load_blob_from_path(path=results_path).fetch())
        elif run_status == RunStatus.ERROR:
            logger.info("Run failed, returning stderr")
            return self.stderr()
        else:
            logger.info(f"Run status: {self.status}, returning stdout")
            return self.stdout()

    def stdout(self) -> str:
        """Read the stdout saved on the system for the Run."""
        stdout_path = self._stdout_path
        logger.info(f"Reading stdout from path: {stdout_path}")

        return self._load_blob_from_path(path=stdout_path).fetch().decode().strip()

    def stderr(self) -> str:
        """Read the stderr saved on the system for the Run."""
        stderr_path = self._stderr_path
        logger.info(f"Reading stderr from path: {stderr_path}")

        return self._load_blob_from_path(stderr_path).fetch().decode().strip()

    def _fn_inputs_path(self) -> str:
        """Path to the pickled inputs used for the function which are saved on the system."""
        return f"{self.folder.path}/{self.INPUTS_FILE}"

    def _fn_result_path(self) -> str:
        """Path to the pickled result for the function which are saved on the system."""
        return f"{self.folder.path}/{self.RESULT_FILE}"

    def _load_blob_from_path(self, path: str):
        """Load a blob from the Run's folder in the specified path. (ex: function inputs, result, stdout, stderr)."""
        return blob(path=path, system=self.folder.system)

    def _register_new_run(self):
        """Log a Run once it's been triggered on the system."""
        self.start_time = self._current_timestamp()
        self.status = RunStatus.RUNNING

        # Write config data for the Run to its config file on the system
        logger.info(f"Registering new Run on system in path: {self.folder.path}")
        self._write_config()

    def _register_fn_run_completion(self, run_status: RunStatus):
        """Update a function based Run's config after its finished running on the system."""
        self.end_time = self._current_timestamp()
        self.status = run_status

        logger.info(f"Registering a completed fn Run with status: {run_status}")
        self._write_config()

    def _register_cmd_run_completion(self, return_codes: list):
        """Update a cmd based Run's config and register its stderr and stdout after running on the system."""
        run_status = RunStatus.ERROR if return_codes[0][0] != 0 else RunStatus.COMPLETED
        self.status = run_status

        logger.info(f"Registering a completed cmd Run with status: {run_status}")
        self._write_config()

        # Write the stdout and stderr of the commands Run to the Run's folder
        self.write(data=return_codes[0][1].encode(), path=self._stdout_path)
        self.write(data=return_codes[0][2].encode(), path=self._stderr_path)

    def _write_config(self, config: dict = None, overwrite: bool = True):
        """Write the Run's config data to the system.

        Args:
            config (Optional[Dict]): Config to write. If none is provided, the Run's config for RNS will be used.
            overwrite (Optional[bool]): Overwrite the config if one is already saved down. Defaults to ``True``.
        """
        config_to_write = config or self.run_config
        logger.info(f"Config to save on system: {config_to_write}")
        self.folder.put(
            {self.RUN_CONFIG_FILE: config_to_write},
            overwrite=overwrite,
            mode="w",
            write_fn=lambda data, f: json.dump(data, f, indent=4),
        )

    def _detect_run_type(self):
        if self.fn_name:
            return RunType.FUNCTION_RUN
        elif self.cmds is not None:
            return RunType.CMD_RUN
        else:
            return RunType.CTX_MANAGER

    def _path_to_config(self) -> str:
        """Path the main folder storing the metadata, inputs, and results for the Run saved on the system."""
        return f"{self.folder.path}/{self.RUN_CONFIG_FILE}"

    def _path_to_file_by_ext(self, ext: str) -> str:
        """Path the file for the Run saved on the system for a provided extension (ex: ``.out`` or ``.err``)."""
        existing_file = self._find_file_path_by_ext(ext=ext)
        if existing_file:
            # If file already exists in file (ex: with function on a Ray cluster this will already be
            # generated for us)
            return existing_file

        path_to_ext = f"{self.folder.path}/{self.name}" + ext
        return path_to_ext

    def _convert_symlink_to_file(self, path: str):
        """If the system is a Cluster and the file path is a symlink, convert it to a regular file.
        This is necessary to allow for copying of the file between systems (ex: cluster --> s3 or cluster --> local)."""
        if isinstance(self.folder.system, Cluster):
            status_codes: list = self.folder.system.run(
                [f"test -h {path} && echo True || echo False"], stream_logs=True
            )
            if status_codes[0][1].strip() == "True":
                # If it's a symlink convert it to a regular file
                self.folder.system.run(
                    [f"cp --remove-destination `readlink {path}` {path}"]
                )

    def _find_file_path_by_ext(self, ext: str) -> Union[str, None]:
        """Get the file path by provided extension. Needed when loading the stdout and stderr files associated
        with a particular run."""
        try:
            folder_contents: list = self.folder.ls(sort=True)
        except FileNotFoundError:
            return None

        files_with_ext = self._filter_files_by_ext(folder_contents, ext)
        if not files_with_ext:
            # No .out / .err file already created in the logs folder for this Run
            return None

        # Return the most recent file with this extension
        return files_with_ext[0]

    def _register_upstream_artifact(self, artifact_name: str):
        """Track a Runhouse object loaded in the Run's context manager. This object's name
        will be saved to the upstream artifact registry of the Run's config."""
        if artifact_name not in self.upstream_artifacts:
            self.upstream_artifacts.append(artifact_name)

    def _register_downstream_artifact(self, artifact_name: str):
        """Track a Runhouse object saved in the Run's context manager. This object's name
        will be saved to the downstream artifact registry of the Run's config."""
        if artifact_name not in self.downstream_artifacts:
            self.downstream_artifacts.append(artifact_name)

    @staticmethod
    def _current_timestamp():
        return str(log_timestamp())

    @staticmethod
    def _filter_files_by_ext(files: list, ext: str):
        return list(filter(lambda x: x.endswith(ext), files))

    @staticmethod
    def _delete_existing_run(folder_path, folder_system: str):
        """Delete existing Run on the system before a new one is created."""
        existing_folder = folder_factory(
            path=folder_path,
            system=folder_system,
        )

        existing_folder.rm()

    @staticmethod
    def _create_new_run_name(name: str = None) -> str:
        """Name of the Run's parent folder which contains the Run's data (config, stdout, stderr, etc).
        If a name is provided, prepend that to the current timestamp to complete the folder name."""
        timestamp_key = f"{datetime.now().strftime('%Y%m%d_%H%M%S')}"
        if name is None:
            return timestamp_key
        return f"{name}_{timestamp_key}"

    @staticmethod
    def _load_run_config(folder: Folder) -> dict:
        """Load the Run config file saved for the Run in its dedicated folder on the system ."""
        try:
            return json.loads(folder.get(Run.RUN_CONFIG_FILE))
        except FileNotFoundError:
            return {}

    @staticmethod
    def _base_cluster_folder_path(name: str):
        """Path to the base folder for this Run on a cluster."""
        return f"{obj_store.LOGS_DIR}/{name}"

    @staticmethod
    def _base_local_folder_path(name: str):
        """Path to the base folder for this Run on a local system."""
        return f"{Run.LOCAL_RUN_PATH}/{name}"


def run(
    name: str = None,
    path: str = None,
    system: Union[str, Cluster] = None,
    data_config: dict = None,
    dryrun: bool = False,
    **kwargs,
) -> Union["Run", None]:
    """Constructs a Run object.

    Args:
        name (Optional[str]): Name of the Run to load.
        path (Optional[str]): Path to the Run's dedicated folder on the system where the Run lives.
        system (Optional[str or Cluster]): File system or cluster name where the Run lives.
            If providing a file system this must be one of:
            [``file``, ``github``, ``sftp``, ``ssh``, ``s3``, ``gs``, ``azure``].
            We are working to add additional file system support.
        data_config (Optional[Dict]): The data config to pass to the underlying fsspec handler for the folder.
        dryrun (bool): Whether to create the Blob if it doesn't exist, or load a Blob object as a dryrun.
            (Default: ``False``)
        **kwargs: Optional kwargs for the Run.

    Returns:
        Run: The loaded Run object.
    """
    if name and not any([path, system, data_config, kwargs]):
        # Try reloading existing Run from RNS
        return Run.from_name(name, dryrun=dryrun)

    if name and path is None:
        path = (
            Run._base_cluster_folder_path(name=name)
            if isinstance(system, Cluster)
            else Run._base_local_folder_path(name=name)
        )

    system = _get_cluster_from(
        system or _current_cluster(key="config") or Folder.DEFAULT_FS, dryrun=dryrun
    )

    run_obj = Run(
        name=name,
        path=path,
        system=system,
        data_config=data_config,
        dryrun=dryrun,
        **kwargs,
    )

    return run_obj
